#=========================================================================
# VerilogPlaceholderPass.py
#=========================================================================
# For each placeholder in the component hierarchy, set up default values,
# check if all configs are valid, and pickle the specified Verilog
# source files.
#
# Author : Peitian Pan
# Date   : Jan 27, 2020
import configparser
import inspect
import os
import re
import sys
from textwrap import dedent

from pymtl3 import MetadataKey
from pymtl3.passes.backends.verilog.util.utility import (
    gen_mapped_ports,
    get_component_unique_name,
)
from pymtl3.passes.errors import InvalidPassOptionValue
from pymtl3.passes.PlaceholderConfigs import expand
from pymtl3.passes.PlaceholderPass import PlaceholderPass
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt


class VerilogPlaceholderPass( PlaceholderPass ):

  # Placeholder pass public pass data

  #: A dict that maps the module parameters to their values.
  #:
  #: Type: ``{ str : int }``; input
  #:
  #: Default value: ``{}``
  params     = MetadataKey(dict)

  #: A dict that maps the PyMTL port name to the external source port name.
  #:
  #: Type: ``{ port : str }``; input
  #:
  #: Default value: ``{}``
  port_map   = MetadataKey(dict)

  #: Top level module name in the external source file.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: PyMTL component class name
  top_module = MetadataKey(str)

  #: Path to the external source file.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: <top_module>.v
  src_file   = MetadataKey(str)

  #: List of Verilog source file paths.
  #:
  #: Type: ``[str]``; input
  #:
  #: Default value: []
  v_flist    = MetadataKey(list)

  #: List of Verilog library file paths. These files will be added to the
  #: beginning of the pikcling result.
  #:
  #: Type: ``[str]``; input
  #:
  #: Default value: []
  v_libs     = MetadataKey(list)

  #: List of Verilog include directory paths.
  #:
  #: Type: ``[str]``; input
  #:
  #: Default value: []
  v_include  = MetadataKey(list)

  #: Separator string used by name-mangling of interfaces and arrays.
  #: For example, with the default value, ``s.ifc.msg`` will be mangled to ``ifc_msg``.
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ``'_'``
  separator  = MetadataKey(str)

  @staticmethod
  def get_placeholder_config():
    from pymtl3.passes.backends.verilog.VerilogPlaceholderConfigs import (
        VerilogPlaceholderConfigs,
    )
    return VerilogPlaceholderConfigs

  # Override
  def visit_placeholder( s, m ):
    c = s.__class__
    super().visit_placeholder( m )
    irepr = RTLIRGetter(cache=False).get_component_ifc_rtlir( m )

    s.setup_default_configs( m, irepr )
    cfg = m.get_metadata( c.placeholder_config )

    if cfg.enable:
      s.check_valid( m, cfg, irepr )
      s.pickle( m, cfg, irepr )

  def setup_default_configs( s, m, irepr ):
    c = s.__class__
    cfg = m.get_metadata( c.placeholder_config )

    if cfg.enable:
      # If top_module is unspecified, infer it from the component and its
      # parameters. Note we need to make sure the infered top_module matches
      # the default translation result.
      if not cfg.top_module:
        cfg.top_module = irepr.get_name()

      # If the placeholder has parameters, use the mangled unique component
      # name. Otherwise use {class_name}_noparam to avoid duplicated defs.
      has_params = bool( irepr.get_params() ) or bool( cfg.params )
      if has_params:
        cfg.pickled_top_module = get_component_unique_name( irepr )
      else:
        cfg.pickled_top_module = f"{irepr.get_name()}_noparam"

      # Only try to infer the name of Verilog source file if both
      # flist and the source file are not specified.
      if not cfg.src_file and not cfg.v_flist:
        parent_dir = os.path.dirname(inspect.getfile(m.__class__))
        cfg.src_file = f"{parent_dir}{os.sep}{cfg.top_module}.v"

      # What is the original file/flist of the pickled source file?
      if cfg.src_file:
        cfg.pickled_orig_file = cfg.src_file
      else:
        cfg.pickled_orig_file = cfg.v_flist

      # The unique placeholder name
      cfg.orig_comp_name = get_component_unique_name( irepr )

      # The `ifdef dependency guard is a function of the placeholder
      # class name
      cfg.dependency_guard_symbol = m.__class__.__name__.upper()

      # The `ifdef placeholder guard is a function of the placeholder
      # wrapper name
      cfg.wrapper_guard_symbol = cfg.pickled_top_module.upper()

      # Scan through src_file to check for clk and reset
      if cfg.is_default('has_clk'):
        cfg.has_clk = s._has_pin(cfg.src_file, 'input', 1, 'clk')

      if cfg.is_default('has_reset'):
        cfg.has_reset = s._has_pin(cfg.src_file, 'input', 1, 'reset')

      # By default, the separator of placeholders is single underscore
      cfg.separator = '_'

      # Look for pymtl.ini starting from the directory that includes the def
      # of class m.__class__
      auto_prefix = s._get_auto_prefix( m )

      # Apply auto_prefix to top_module
      cfg.top_module = auto_prefix + cfg.top_module

  def check_valid( s, m, cfg, irepr ):
    pmap, src, flist, include = \
        cfg.port_map, cfg.src_file, cfg.v_flist, cfg.v_include

    # Check params
    for param_name, value in cfg.params.items():
      if not isinstance( value, int ):
        raise InvalidPassOptionValue("params", cfg.params, cfg.Pass.__name__,
            f"non-integer parameter {param_name} is not supported yet!")

    # Check port map
    # TODO: this should be based on RTLIR
    for port in pmap.keys():
      if port.get_host_component() is not m:
        raise InvalidPassOptionValue("port_map", pmap, cfg.Pass.__name__,
          f"Port {port} does not exist in component {irepr.get_name()}!")

    # Check src_file
    if cfg.src_file and not os.path.isfile(expand(cfg.src_file)):
      raise InvalidPassOptionValue("src_file", cfg.src_file, cfg.Pass.__name__,
          'src_file should be a file path!')

    if cfg.v_flist:
      raise InvalidPassOptionValue("v_flist", cfg.v_flist, cfg.Pass.__name__,
          'Placeholders backed by Verilog flist are not supported yet!')

    # Check v_flist
    if cfg.v_flist and not os.path.isfile(expand(cfg.v_flist)):
      raise InvalidPassOptionValue("v_flist", cfg.v_flist, cfg.Pass.__name__,
          'v_flist should be a file path!')

    # Check v_include
    if cfg.v_include:
      for include in cfg.v_include:
        if not os.path.isdir(expand(include)):
          raise InvalidPassOptionValue("v_include", cfg.v_include, cfg.Pass.__name__,
              'v_include should be an array of dir paths!')

    # Check if the top module name appears in the file
    if cfg.src_file:
      found = False
      with open(cfg.src_file) as src_file:
        for line in src_file.readlines():
          if cfg.top_module in line:
            found = True
            break
      if not found:
        raise InvalidPassOptionValue("top_module", cfg.top_module, cfg.Pass.__name__,
            f'cannot find top module {cfg.top_module} in source file {cfg.src_file}.\n'
            f'Please make sure you have specified the correct top module name through '
            f'the VerilogPlaceholderPass.top_module pass data name!')

  def pickle( s, m, cfg, irepr ):
    pickled_dependency    = s._get_v_lib_files( m, cfg, irepr )
    pickled_dependency   += s._get_dependent_verilog_modules( m, cfg, irepr )
    pickled_wrapper, tplt = s._gen_verilog_wrapper( m, cfg, irepr )

    pickled_dependency_source = (
        f"//***********************************************************\n"
        f"// Pickled source file of placeholder {cfg.orig_comp_name}\n"
        f"//***********************************************************\n"
        f"\n"
        f"//-----------------------------------------------------------\n"
        f"// Dependency of placeholder {m.__class__.__name__}\n"
        f"//-----------------------------------------------------------\n"
        f"\n"
        f"`ifndef {cfg.dependency_guard_symbol}\n"
        f"`define {cfg.dependency_guard_symbol}\n"
        f"\n"
        f"{pickled_dependency}"
        f"\n"
        f"`endif /* {cfg.dependency_guard_symbol} */\n"
    )

    pickled_wrapper_source_tplt = (
        f"\n"
        f"//-----------------------------------------------------------\n"
        f"// Wrapper of placeholder {cfg.orig_comp_name}\n"
        f"//-----------------------------------------------------------\n"
        f"\n"
        f"`ifndef {cfg.wrapper_guard_symbol}\n"
        f"`define {cfg.wrapper_guard_symbol}\n"
        f"\n"
        f"{{pickled_wrapper}}\n"
        f"\n"
        f"`endif /* {cfg.wrapper_guard_symbol} */\n"
    )

    pickled_wrapper_source = pickled_wrapper_source_tplt.format(pickled_wrapper=pickled_wrapper)
    pickled_source = pickled_dependency_source + pickled_wrapper_source

    cfg.pickled_wrapper_template = pickled_wrapper_source_tplt.format(pickled_wrapper=tplt)
    cfg.pickled_wrapper_nlines   = len(pickled_wrapper_source.split('\n'))-1

    cfg.pickled_source = pickled_source

  def _get_v_lib_files( s, m, cfg, irepr ):
    orig_comp_name = cfg.orig_comp_name
    tplt = dedent(
        """\
            // The source code below are included because they are specified
            // as the v_libs Verilog placeholder option of component {orig_comp_name}.

            // If you get a duplicated def error from files included below, please
            // make sure they are included either through the v_libs option or the
            // explicit `include statement in the Verilog source code -- if they
            // appear in both then they will be included twice!

            {v_libs}
            // End of all v_libs files for component {orig_comp_name}

        """
    )
    if cfg.v_libs:
      v_libs = s._import_sources( cfg, cfg.v_libs )
    else:
      v_libs = ""

    return tplt.format(**locals())

  def _get_dependent_verilog_modules( s, m, cfg, irepr ):
    return s._import_sources( cfg, [cfg.src_file] )

  def _gen_verilog_wrapper( s, m, cfg, irepr ):
    # By default use single underscore as separator
    rtlir_ports = gen_mapped_ports( m, cfg.port_map, cfg.has_clk, cfg.has_reset, '_' )

    all_port_names = list(map(lambda x: x[1], rtlir_ports))

    if not cfg.params:
      parameters = irepr.get_params()
    else:
      parameters = cfg.params.items()

    # Port definitions of wrapper
    ports = []
    for idx, (_, name, p, _) in enumerate(rtlir_ports):
      if name:
        if isinstance(p, rt.Array):
          n_dim = p.get_dim_sizes()
          s_dim = ''.join([f'[0:{idx-1}]' for idx in n_dim])
          p = p.get_next_dim_type()
        else:
          s_dim = ''
        ports.append(
            f"  {p.get_direction()} logic [{p.get_dtype().get_length()}-1:0]"\
            f" {name} {s_dim}{'' if idx == len(rtlir_ports)-1 else ','}"
        )

    # The wrapper has to have an unused clk port to make verilator
    # VCD tracing work.
    if 'clk' not in all_port_names:
      ports.insert( 0, '  input logic clk,' )

    if 'reset' not in all_port_names:
      ports.insert( 0, '  input logic reset,' )

    # Parameters passed to the module to be parametrized
    params = [
      f"    .{param}( {val} ){'' if idx == len(parameters)-1 else ','}"\
      for idx, (param, val) in enumerate(parameters)
    ]

    # Connections between top module and inner module
    connect_ports = [
      f"    .{name}( {name} ){'' if idx == len(rtlir_ports)-1 else ','}"\
      for idx, (_, name, p, _) in enumerate(rtlir_ports) if name
    ]

    lines = [
      f"module {cfg.pickled_top_module}",
      "(",
    ] + ports + [
      ");",
      f"  {cfg.top_module}",
      "  #(",
    ] + params + [
      "  ) v",
      "  (",
    ] + connect_ports + [
      "  );",
      "endmodule",
    ]

    template_lines = [
      "module {top_module_name}",
      "(",
    ] + ports + [
      ");",
      f"  {cfg.top_module}",
      "  #(",
    ] + params + [
      "  ) v",
      "  (",
    ] + connect_ports + [
      "  );",
      "endmodule",
    ]

    return '\n'.join( line for line in lines ), '\n'.join( line for line in template_lines )

  #-----------------------------------------------------------------------
  # _has_pin
  #-----------------------------------------------------------------------

  def _has_pin( s, src_file, direction, nbits, pin ):
    trim = lambda s: ''.join(s.split())
    targets = [
        f'{direction}logic[{nbits-1}:0]{pin}',
        f'{direction}[{nbits-1}:0]{pin}'
    ]
    if nbits == 1:
      targets.append( f'{direction}logic{pin}' )
      targets.append( f'{direction}{pin}' )
    try:
      with open(src_file) as fd:
        for line in fd.readlines():
          if any(map(lambda x: x in trim(line), targets)):
            return True
      return False
    except OSError:
      return False

  #-----------------------------------------------------------------------
  # _get_auto_prefix
  #-----------------------------------------------------------------------

  def _get_auto_prefix( s, m ):
    parent_dir = cwd = os.path.dirname(inspect.getfile(m.__class__))
    while cwd != '/':
      if os.path.exists(f"{cwd}{os.path.sep}pymtl.ini"):
        break
      cwd = os.path.dirname(cwd)

    if cwd == '/': return ''

    pymtl_config = configparser.ConfigParser()
    pymtl_config.read(f"{cwd}{os.path.sep}pymtl.ini")

    if 'placeholder' in pymtl_config and \
       'auto_prefix' in pymtl_config['placeholder'] and \
       pymtl_config.getboolean( 'placeholder', 'auto_prefix' ):
      if cwd == parent_dir:
        # Just return the base name
        return f"{os.path.basename(parent_dir)}_"
      else:
        # Return an underscore-connected prefix that includes all
        # directories along the path
        return f"{parent_dir[len(cwd)+1:].replace(os.path.sep,'_')}_"
    else:
      return ''

  #-----------------------------------------------------------------------
  # import_sources
  #-----------------------------------------------------------------------
  # The right way to do this is to use a recursive function like I have
  # done below. This ensures that files are inserted into the output stream
  # in the correct order. -cbatten

  # Regex to extract verilog filenames from `include statements

  _include_re = re.compile( r'"(?P<filename>[\w/\.-]*)"' )

  def _output_verilog_file( s, include_path, verilog_file ):
    code = ""
    with open(verilog_file) as fp:

      short_verilog_file = verilog_file
      if verilog_file.startswith( include_path+"/" ):
        short_verilog_file = verilog_file[len(include_path+"/"):]

      code += '`line 1 "{}" 0\n'.format( short_verilog_file )

      line_num = 0
      for line in fp:
        line_num += 1
        if '`include' in line:
          re_result = s._include_re.search( line )
        else:
          re_result = None

        if re_result:
          filename = re_result.group( 'filename' )
          fullname = os.path.join( include_path, filename )
          code += s._output_verilog_file( include_path, fullname )
          code += '\n'
          code += '`line {} "{}" 0\n'.format( line_num+1, short_verilog_file )
        else:
          code += line
    return code

  def _import_sources( s, cfg, source_list ):
    # Import Verilog source from all Verilog files source_list. This includes
    # any source files specified by `include within those files.

    code = ""

    if not source_list:
      return

    # We will use the first verilog file to find the root of PyMTL project

    first_verilog_file = source_list[0]

    # All verilog includes are relative to the root of the PyMTL project.
    # We identify the root of the PyMTL project by looking for the special
    # pymtl.ini file.

    _path = os.path.dirname( first_verilog_file )
    special_file_found = False
    include_path = os.path.dirname( os.path.abspath( first_verilog_file ) )
    while include_path != "/":
      if os.path.exists( include_path + os.path.sep + "pymtl.ini" ):
        special_file_found = True
        sys.path.insert(0,include_path)
        break
      include_path = os.path.dirname(include_path)

    # Append the user-defined include path to include_path
    # NOTE: the current pickler only supports one include path. If v_include
    # config is present, use it instead.

    if cfg.v_include:
      if len(cfg.v_include) != 1:
        raise InvalidPassOptionValue("v_include", cfg.v_include, cfg.Pass.__name__,
            'the current pickler only supports one user-defined v_include path...')
      include_path = cfg.v_include[0]

    # If we could not find the special .pymtl-python-path file, then assume
    # the include directory is the same as the directory that contains the
    # first verilog file.

    if not special_file_found and not cfg.v_include:
      include_path = os.path.dirname( os.path.abspath( first_verilog_file ) )

    # Regex to extract verilog filenames from `include statements

    s._include_re = re.compile( r'"(?P<filename>[\w/\.-]*)"' )

    # Iterate through all source files and add any `include files to the
    # list of source files to import.

    for source in source_list:
      code += s._output_verilog_file( include_path, source )

    return code



#=========================================================================
# YosysTranslationImportPass.py
#=========================================================================
# Translate and import components in the given hierarchy.
#
# Author : Peitian Pan
# Date   : Aug 6, 2019

from pymtl3.passes.backends.verilog.VerilogTranslationImportPass import (
    VerilogTranslationImportPass,
)

from .import_.YosysVerilatorImportPass import YosysVerilatorImportPass
from .translation.YosysTranslationPass import YosysTranslationPass


class YosysTranslationImportPass( VerilogTranslationImportPass ):

  @staticmethod
  def get_translation_pass():
    return YosysTranslationPass

  @staticmethod
  def get_import_pass():
    return YosysVerilatorImportPass



#=========================================================================
# VerilogTranslationImportPass.py
#=========================================================================
# Translate and import components in the given hierarhcy.
#
# Author : Peitian Pan
# Date   : Aug 6, 2019

from pymtl3.dsl import MetadataKey
from pymtl3.passes.BasePass import BasePass

from .import_.VerilogVerilatorImportConfigs import VerilogVerilatorImportConfigs
from .import_.VerilogVerilatorImportPass import VerilogVerilatorImportPass
from .translation.VerilogTranslationConfigs import VerilogTranslationConfigs
from .translation.VerilogTranslationPass import VerilogTranslationPass
from .VerilogPlaceholder import VerilogPlaceholder
from .VerilogPlaceholderConfigs import VerilogPlaceholderConfigs
from .VerilogPlaceholderPass import VerilogPlaceholderPass


class VerilogTranslationImportPass( BasePass ):

  #: Enable translation-import on the component.
  #:
  #: Type: ``bool``; input
  #:
  #: Default value: ``False``
  enable = MetadataKey(bool)

  def __call__( s, top ):
    """Translate-import the PyMTL component hierarhcy rooted at ``top``."""
    c = s.__class__
    s.top = top
    s.traverse_hierarchy( top )
    top.apply( c.get_translation_pass()() )
    s.add_placeholder_marks( top )
    return c.get_import_pass()()( top )

  def traverse_hierarchy( s, m ):
    c = s.__class__

    # Found a subtree that is marked as to be translated and imported
    if ( m.has_metadata( c.enable ) and m.get_metadata( c.enable ) ) or \
      isinstance( m, VerilogPlaceholder ):

      # Make sure the translation pass is enabled
      m.set_metadata( c.get_translation_pass().enable, True )

      # Make sure the import pass is enabled
      m.set_metadata( c.get_import_pass().enable, True )

    else:
      for child in m.get_child_components(repr):
        s.traverse_hierarchy( child )

  def add_placeholder_marks( s, m ):
    c = s.__class__

    # TODO: what we really want is to generate placeholders for components
    # to be translated and imported.
    if ( m.has_metadata( c.enable ) and m.get_metadata( c.enable ) ) or \
      isinstance( m, VerilogPlaceholder ):

      # Component m will look as if it has applied the Placeholder pass.
      # It will have the output pass data placeholder_config

      placeholder_pass = c.get_placeholder_pass()
      translation_pass = c.get_translation_pass()
      m.set_metadata( placeholder_pass.enable, True )

      if m.has_metadata( placeholder_pass.placeholder_config ):
        placeholder_config = m.get_metadata( placeholder_pass.placeholder_config )
      else:
        placeholder_config = c.get_placeholder_config()( m )

        if not isinstance( m, VerilogPlaceholder ):
          # If m is not a placeholder, we need to populate the v_include
          # placeholder config from all submodules into m.
          placeholder_config.v_include = list(c.get_hierarchy_v_include(m))

      placeholder_config.pickled_source_file = \
          m.get_metadata( translation_pass.translated_filename )
      placeholder_config.pickled_top_module = \
          m.get_metadata( translation_pass.translated_top_module )
      m.set_metadata( placeholder_pass.placeholder_config, placeholder_config )

    else:
      for child in m.get_child_components(repr):
        s.add_placeholder_marks( child )

  @staticmethod
  def get_translation_pass():
    return VerilogTranslationPass

  @staticmethod
  def get_import_pass():
    return VerilogVerilatorImportPass

  @staticmethod
  def get_placeholder_pass():
    return VerilogPlaceholderPass

  @staticmethod
  def get_placeholder_config():
    return VerilogPlaceholderConfigs

  # This recursive method needs to be a class method
  @classmethod
  def get_hierarchy_v_include( c, m ):
    all_v_includes = set()
    ph_config_key = c.get_placeholder_pass().placeholder_config
    if isinstance( m, VerilogPlaceholder ) and m.has_metadata( ph_config_key ):
      ph_config = m.get_metadata( ph_config_key )
      for v_include in ph_config.v_include:
        all_v_includes.add( v_include )
    else:
      for child in m.get_child_components(repr):
        all_v_includes |= c.get_hierarchy_v_include( child )
    return all_v_includes



#=========================================================================
# BehavioralTranslatorL4.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L4 behavioral translator."""

from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL4Pass import (
    BehavioralRTLIRGenL4Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL4Pass import (
    BehavioralRTLIRTypeCheckL4Pass,
)

from .BehavioralTranslatorL3 import BehavioralTranslatorL3


class BehavioralTranslatorL4( BehavioralTranslatorL3 ):

  #-----------------------------------------------------------------------
  # _gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def _gen_behavioral_trans_metadata( s, m ):
    m.apply( BehavioralRTLIRGenL4Pass( s.tr_top ) )
    m.apply( BehavioralRTLIRTypeCheckL4Pass( s.tr_top ) )
    s.behavioral.rtlir[m] = \
        m.get_metadata( BehavioralRTLIRGenL4Pass.rtlir_upblks )
    s.behavioral.freevars[m] =\
        m.get_metadata( BehavioralRTLIRTypeCheckL4Pass.rtlir_freevars )
    s.behavioral.tmpvars[m] =\
        m.get_metadata( BehavioralRTLIRTypeCheckL4Pass.rtlir_tmpvars )



"""
========================================================================
VcdGenerationPass.py
========================================================================

Author : Shunning Jiang, Yanghui Ou, Peitian Pan
Date   : Sep 8, 2019
"""

import time
from collections import defaultdict

from pymtl3.datatypes import Bits, concat
from pymtl3.dsl import Const, MetadataKey
from pymtl3.passes.BasePass import BasePass
from pymtl3.passes.errors import PassOrderError


class VcdGenerationPass( BasePass ):

  # VcdGenerationPass pass public pass data

  #: vcd file name
  #:
  #: Type: ``str``; input
  #:
  #: Default value: ""
  vcd_file_name = MetadataKey(str)

  vcd_func = MetadataKey()

  def __call__( self, top ):
    if top.has_metadata( self.vcd_file_name ):
      vcd_file_name = top.get_metadata( self.vcd_file_name )

      if vcd_file_name is not None:
        assert not top.has_metadata( self.vcd_func )
        top.set_metadata( self.vcd_func, self.make_vcd_func( top, vcd_file_name ) )

  def make_vcd_func( self, top, vcd_file_name ):
    assert vcd_file_name is not None
    if vcd_file_name != "":
      vcd_file_name = str(vcd_file_name) + ".vcd"
    else:
      vcd_file_name = str(top.__class__.__name__) + ".vcd"

    vcd_file = open( vcd_file_name, "w" )

    # Get vcd timescale

    try:                    vcd_timescale = top.vcd_timescale
    except AttributeError:  vcd_timescale = "10ps"

    # Print vcd header

    print( "$date\n  {}\n$end\n$version\n  PyMTL 3 (Mamba)\n$end\n"
           "$timescale\n {}\n$end\n".format( time.asctime(), vcd_timescale ),
           file=vcd_file )

    # Utility generator to create new symbols for each VCD signal.
    # Code inspired by MyHDL 0.7.
    # Shunning: I just reuse it from pymtl v2

    def _gen_vcd_symbol():

      # Generate a string containing all valid vcd symbol characters
      _codechars = ''.join([chr(i) for i in range(33, 127)])
      _mod       = len(_codechars)

      # Generator logic
      n = 0
      while True:
        q, r = divmod(n, _mod)
        code = _codechars[r]
        while q > 0:
          q, r = divmod(q, _mod)
          code = _codechars[r] + code
        yield code
        n += 1

    vcd_symbols = _gen_vcd_symbol()

    # Preprocess some metadata

    component_signals = defaultdict(set)

    all_components = set()

    # We only collect top level signals, and squash bitstruct into a long
    # bits object
    for x in top._dsl.all_signals:
      if x.is_top_level_signal():
        host = x.get_host_component()
        component_signals[ host ].add( x )

    # We pre-process all nets in order to remove all sliced wires because
    # they belong to a top level wire and we count that wire

    trimmed_value_nets = []
    vcd_clock_net_idx = None

    # FIXME handle the case where the top level signal is in a value net
    for writer, net in top.get_all_value_nets():
      new_net = []
      for x in net:
        if not isinstance(x, Const) and x.is_top_level_signal():
          new_net.append( x )
          if repr(x) == "s.clk":
            # Hardcode clock net because it needs to go up and down
            assert vcd_clock_net_idx is None
            vcd_clock_net_idx = len(trimmed_value_nets)

      if new_net:
        trimmed_value_nets.append( new_net )

    # Generate symbol for existing nets

    net_symbol_mapping = [ next(vcd_symbols) for x in trimmed_value_nets ]
    signal_net_mapping = {}

    for i in range(len(trimmed_value_nets)):
      for x in trimmed_value_nets[i]:
        signal_net_mapping[x] = i

    # Inner utility function to perform recursive descent of the model.
    # Shunning: I mostly follow v2's implementation

    # Vcd file takes a(0) instead of a[0]
    def vcd_mangle_name( name ):
      # signal names with colons in it silently fail gtkwave
      return name.replace('[','(').replace(']',')').replace(':', '__')

    def recurse_models( m, spaces ):
      nonlocal vcd_clock_net_idx, vcd_sim_ncycles

      # Special case the top level "s" to "top"

      my_name = m.get_field_name()
      if my_name == "s":
        my_name = "top"

      # Create a new scope for this module
      print( f"{spaces}$scope module {vcd_mangle_name(my_name)} $end",
             file=vcd_file )

      m_name = repr(m)

      # Define all signals for this model.
      for signal in component_signals[m]:

        # Multiple signals may be collapsed into a single net in the
        # simulator if they are connected. Generate new vcd symbols per
        # net, not per signal as an optimization.

        if signal in signal_net_mapping:
          net_id = signal_net_mapping[signal]
          symbol = net_symbol_mapping[net_id]
        else:
          # We treat this as a new net

          # Check if it's clock. Hardcode clock net
          if repr(signal) == "s.clk":
            assert vcd_clock_net_idx is None
            vcd_clock_net_idx = len(trimmed_value_nets)

          # This is a signal whose connection is not captured by the
          # global net data structure. This might be a sliced signal or
          # a signal updated in an upblk. Creating a new net for it does
          # not hurt functionality.

          trimmed_value_nets.append( [ signal ] )
          signal_net_mapping[signal] = len(signal_net_mapping)
          symbol = next(vcd_symbols)
          net_symbol_mapping.append( symbol )

        # This signal can be a part of an interface so we have to
        # "subtract" host component's name from signal's full name
        # to get the actual name like enq.rdy
        # TODO struct
        signal_name = vcd_mangle_name( repr(signal)[ len(m_name)+1: ] )
        print( f"{spaces}  $var reg {signal._dsl.Type.nbits} {symbol} {signal_name} $end",
               file=vcd_file )

      # Recursively visit all submodels.
      for child in m.get_child_components():
        recurse_models( child, spaces+'  ' )

      print( f"{spaces}$upscope $end", file=vcd_file )

    # Begin recursive descent from the top-level model.
    recurse_models( top, '' )

    # Once all models and their signals have been defined, end the
    # definition section of the vcd and print the initial values of all
    # nets in the design.
    print( "$enddefinitions $end\n", file=vcd_file )

    # last_values is an array of values from the previous cycle

    last_values = [0 for _ in range(len(trimmed_value_nets))]

    for i, net in enumerate(trimmed_value_nets):
      # Convert everything to Bits to get around lack of bit struct support.
      # The first cycle VCD contains the default value
      bin_str = net[0]._dsl.Type().to_bits().bin()

      print( f"b{bin_str} {net_symbol_mapping[i]}", file=vcd_file )

      # Set this to be the last cycle value str
      last_values[i] = bin_str

    # Now we create per-cycle signal value collect functions

    vcd_sim_ncycles = 0

    # Separate clock net from normal nets ahead of time
    clock_symbol = net_symbol_mapping[ vcd_clock_net_idx ]

    net_details = [ ( trimmed_value_nets[i][0], net_symbol_mapping[i] )
                    for i in range(len(trimmed_value_nets))
                      if i != vcd_clock_net_idx ]

    # Flip clock for the first cycle
    print( '\n#0\nb0b1 {}\n'.format( clock_symbol ), file=vcd_file, flush=True )

    # Returns a dump_vcd function that is ready to be appended to _sched.
    # TODO: type check?

    # Adding this 's' argument is for eval to correctly evaluate 's.x'...
    # Python 3 destroys a lot of our hacks .. sigh

    def dump_vcd_inner( s ):
      nonlocal vcd_sim_ncycles

      for i, (signal, symbol) in enumerate( net_details ):

        # If we encounter a BitStruct then dump it as a concatenation of
        # all fields.
        # TODO: treat each field in a BitStruct as a separate signal?

        try:
          net_bits_bin = eval(repr(signal)).to_bits()
        except Exception as e:
          raise TypeError(f'{e}\n - {signal} becomes another type. Please check your code.')

        net_bits_bin_str = net_bits_bin.bin()
        # `last_value` is the string form of a Bits object in binary
        # e.g. '0b000' == Bits3(0).bin()
        # We store strings instead of values ...
        if last_values[i] != net_bits_bin_str:
          last_values[i] = net_bits_bin_str
          print( f'b{net_bits_bin_str} {symbol}', file=vcd_file )

      # Flop clock at the end of cycle
      next_neg_edge = 100 * vcd_sim_ncycles + 50
      print( f'\n#{next_neg_edge}\nb0b0 {clock_symbol}', file=vcd_file )

      # Flip clock of the next cycle
      next_pos_edge = next_neg_edge + 50
      print( f'#{next_pos_edge}\nb0b1 {clock_symbol}\n', file=vcd_file, flush=True )
      vcd_sim_ncycles += 1

    def gen_dump_vcd( s ):
      def dump_vcd():
        dump_vcd_inner( s )
      return dump_vcd

    return gen_dump_vcd( top )



#=========================================================================
# VStructuralTranslatorL4.py
#=========================================================================
"""Provide SystemVerilog structural translator implementation."""

from textwrap import dedent

from pymtl3.passes.backends.generic.structural.StructuralTranslatorL4 import (
    StructuralTranslatorL4,
)
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRGetter
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.RTLIRPass import RTLIRPass

from ...util.utility import make_indent, pretty_concat
from ...VerilogPlaceholder import VerilogPlaceholder
from .VStructuralTranslatorL3 import VStructuralTranslatorL3


class VStructuralTranslatorL4(
    VStructuralTranslatorL3, StructuralTranslatorL4 ):

  #-----------------------------------------------------------------------
  # Declarations
  #-----------------------------------------------------------------------

  def rtlir_tr_subcomp_port_decls( s, _port_decls ):
    return _port_decls

  def rtlir_tr_subcomp_port_decl( s, m, c_id, c_rtype, c_array_type, port_id,
      port_rtype, port_dtype, port_array_type ):
    obj = c_rtype.obj
    if obj.has_metadata(s._placeholder_pass.placeholder_config):
      pmap = obj.get_metadata(s._placeholder_pass.placeholder_config).get_port_map()
    else:
      pmap= lambda x: x
    direction = port_rtype.get_direction()
    if direction == 'input':
      direction += ' '
    vname = port_id
    port_full_name = f"{obj}.{port_id}"
    ph_id = vname if pmap(port_full_name) == port_full_name else pmap(port_full_name)
    return {
        'direction' : direction,
        'data_type' : port_dtype['data_type'],
        'packed_type' : port_dtype['packed_type'],
        'id' : port_id,
        'ph_id' : ph_id,
        'unpacked_type' : port_array_type['unpacked_type'],
    }

  def rtlir_tr_subcomp_ifc_port_decls( s, _ifc_port_decls ):
    return sum(_ifc_port_decls, [])

  def rtlir_tr_subcomp_ifc_port_decl( s, m, c_id, c_rtype, c_array_type,
      ifc_id, ifc_rtype, ifc_array_type, port_id, port_rtype, port_array_type ):
    if isinstance( port_rtype, rt.Port ):
      obj = c_rtype.obj
      if obj.has_metadata(s._placeholder_pass.placeholder_config):
        pmap = obj.get_metadata(s._placeholder_pass.placeholder_config).get_port_map()
      else:
        pmap = lambda x: x
      vname = f'{ifc_id}__{port_id}'
      pyname = vname.replace('__', '.')
      # `pmap` is a function that maps the _full name_ of a port object to its
      # Verilog name. `pyname` is simply a name in the local scope. We need to
      # build up the full name by concatenating `str(obj)` and `pyname`.
      port_full_name = f"{obj}.{pyname}"
      ph_id = vname if pmap(port_full_name) == port_full_name else pmap(port_full_name)
      port_dtype = s.rtlir_data_type_translation( m, port_rtype.get_dtype() )
      direction = port_rtype.get_direction()
      if direction == 'input':
        direction += ' '
      return [{
          'direction' : direction,
          'data_type' : port_dtype['data_type'],
          'packed_type' : port_dtype['packed_type'],
          'id' : vname,
          'ph_id' : ph_id,
          'unpacked_type' : ifc_array_type['unpacked_type']+port_array_type['unpacked_type'],
      }]
    else:
      # Nested interface
      ret = []
      all_properties = port_rtype.get_all_properties_packed()
      for _port_id, _port_rtype in all_properties:
        if isinstance(_port_rtype, rt.Array):
          _port_array_rtype = _port_rtype
          _port_rtype = _port_rtype.get_sub_type()
        else:
          _port_array_rtype = None
          _port_rtype = _port_rtype

        # Combine the array_type of two interfaces into one
        combined_ifc_array_type = {
            'def' : port_array_type['def'],
            'unpacked_type' : ifc_array_type['unpacked_type'] + port_array_type['unpacked_type'],
            'n_dim' : ifc_array_type['n_dim'] + port_array_type['n_dim'],
        }

        ret += s.rtlir_tr_subcomp_ifc_port_decl( m,
                  # Component: nested interface does not change the component
                  c_id, c_rtype, c_array_type,
                  # Interface: nested interface appends its id and array_type
                  f'{ifc_id}__{port_id}', port_rtype, combined_ifc_array_type,
                  # Port: use the id, rtype, and array_type of the port
                  _port_id, _port_rtype, s.rtlir_tr_unpacked_array_type(_port_array_rtype))
      return ret

  def rtlir_tr_subcomp_ifc_decls( s, _ifc_decls ):
    return sum(_ifc_decls, [])

  def rtlir_tr_subcomp_ifc_decl( s, m, c_id, c_rtype, c_array_type,
      ifc_id, ifc_rtype, ifc_array_type, ports ):
    return ports

  def rtlir_tr_subcomp_decls( s, subcomps ):
    subcomp_decls = sum( subcomps, [] )
    return '\n\n'.join( subcomp_decls )

  def rtlir_tr_subcomp_decl( s, m, c_id, c_rtype, c_array_type, port_conns, ifc_conns ):

    def pretty_comment( string ):
      comments = [
          '  //-------------------------------------------------------------',
         f'  // {string}',
          '  //-------------------------------------------------------------',
      ]
      return '\n'.join(comments)

    def gen_subcomp_array_decl( c_id, port_conns, ifc_conns, n_dim, c_n_dim ):
      nonlocal m, s
      tplt = dedent(
          """\
            {c_name} {c_id}
            (
          {port_conn_decls}
            );""")
      if not n_dim:
        # Get the object from the hierarchy
        _n_dim = list(int(num_str) for num_str in c_n_dim.split('__') if num_str)
        attr = c_id + ''.join(f'[{dim}]' for dim in _n_dim)
        obj = eval(f'm.{attr}')
        # Get the translated component name
        obj_c_rtype = s.tr_top.get_metadata(RTLIRPass.rtlir_getter).get_rtlir(obj)
        _c_name = s.rtlir_tr_component_unique_name(obj_c_rtype)

        # Check to see if explicit_module_name is present
        from pymtl3.passes.backends.verilog.translation.VerilogTranslationPass import (
            VerilogTranslationPass,
        )
        if obj.has_metadata( VerilogTranslationPass.explicit_module_name ):
          subcomp_explicit_name = obj.get_metadata( VerilogTranslationPass.explicit_module_name )
        else:
          subcomp_explicit_name = ''

        if isinstance(obj, VerilogPlaceholder):
          c_name = obj.get_metadata( s._placeholder_pass.placeholder_config ).pickled_top_module
        elif subcomp_explicit_name:
          # If someone sets explicit_module_name, we need to honor that config
          c_name = subcomp_explicit_name
        else:
          c_name = _c_name

        orig_c_id = c_id
        c_id = c_id + c_n_dim

        # Generate correct connections
        port_conn_decls = []
        unpacked_str = ''.join([f'[{i}]' for i in _n_dim])

        no_clk   = s.structural.component_no_synthesis_no_clk[obj]
        no_reset = s.structural.component_no_synthesis_no_reset[obj]

        # Move clk and reset ports to the front if no_clk/no_reset is true

        conns           = []
        clk_reset_conns = []

        for dscp in port_conns + ifc_conns:
            port_name = dscp['id']
            if (port_name == 'clk' and no_clk) or (port_name == 'reset' and no_reset):
                clk_reset_conns.append(dscp)
            else:
                conns.append(dscp)

        conns = clk_reset_conns + conns

        for i, dscp in enumerate(conns):
          comma = ',\n' if i != len(conns)-1 else ''
          port_name = dscp['id']
          ph_port_name = dscp['ph_id']
          port_wire = f"{orig_c_id}__{dscp['id']}{unpacked_str}"
          if (port_name == 'clk' and no_clk) or (port_name == 'reset' and no_reset):
            comma = ',\n' if i != len(conns)-1 else '\n'
            newline = '\n' if i != len(conns)-1 else ''
            port_conn_decls.append("`ifndef SYNTHESIS\n")
            port_conn_decls.append(f".{ph_port_name}( {port_wire} ){comma}")
            port_conn_decls.append(f"`endif{newline}")
          else:
            port_conn_decls.append(f".{ph_port_name}( {port_wire} ){comma}")

        make_indent( port_conn_decls, 2 )
        port_conn_decls = ''.join(port_conn_decls)
        return [ tplt.format( **locals() ) ]

      else:
        return sum( [gen_subcomp_array_decl( c_id,
            port_conns, ifc_conns, n_dim[1:], c_n_dim+'__'+str(idx) ) \
                for idx in range( n_dim[0] )], [] )

    # If `c_array_type` is not None we need to impelement an array of
    # components, each with their own connections for the ports.

    # Generate wire declarations for all ports
    defs = []

    for dscp in port_conns + ifc_conns:
      defs.append(pretty_concat(dscp['data_type'], dscp['packed_type'],
        f"{c_id}__{dscp['id']}", f"{c_array_type['unpacked_type']}{dscp['unpacked_type']}", ';'))

    make_indent( defs, 1 )
    defs = ['\n'.join(defs)]

    n_dim = c_array_type['n_dim']
    subcomps = gen_subcomp_array_decl( c_id, port_conns, ifc_conns, n_dim, '' )

    return [pretty_comment(f"Component {c_id}{c_array_type['unpacked_type']}")] + \
           defs + subcomps + \
           [pretty_comment(f"End of component {c_id}{c_array_type['unpacked_type']}")]

  #-----------------------------------------------------------------------
  # Signal operations
  #-----------------------------------------------------------------------

  def rtlir_tr_component_array_index( s, base_signal, index, status ):
    s._rtlir_tr_unpacked_q.append( index )
    return base_signal

  def rtlir_tr_subcomp_attr( s, base_signal, attr, status ):
    return s._rtlir_tr_process_unpacked(
              f'{base_signal}__{attr}',
              f'{base_signal}__{attr}{{}}',
              status, ('status') )



#=========================================================================
# pytest_pymtl3.py
#=========================================================================
# PyMTL3 pytest plugin. Reused the conftest.py file from pymtl3 repo.
# With the current setup, you can add `pytest_plugins = "pytest_pymtl3"`
# to obtain access to the pytest fixtures defined below.
#
# Author : Peitian Pan
# Date   : March 30, 2020

import os

import pytest

if 'CI' in os.environ:
  # Set up the CI hypothesis profile which limits the max number of tries
  # The 'CI' profile will be specified through the testing command of the
  # CI script.
  from hypothesis import settings
  settings.register_profile("CI", max_examples=10)

#-------------------------------------------------------------------------
# pytest hooks and fixtures
#-------------------------------------------------------------------------

def pytest_addoption(parser):
  group = parser.getgroup("pytest-pymtl3")
  group.addoption( "--test-verilog", dest="test_verilog", action="store",
                    default='', nargs='?', const='zeros',
                    help="run verilog translation" )
  group.addoption( "--dump-textwave", dest="dump_textwave", action="store_true",
                    default=None, help="dump text waveform for each test" )
  group.addoption( "--dump-vcd", dest="dump_vcd", action="store_true",
                    default=None, help="dump vcd for each test" )
  group.addoption( "--dump-vtb", dest="dump_vtb", action="store_true",
                    default=None, help="dump verilog test bench for each test" )
  group.addoption( "--max-cycles", dest="max_cycles", action="store",
                    default=None, help="max cycles of simulation" )

@pytest.fixture
def cmdline_opts( request ):
  """PyMTL options parsed from pytest commandline options."""
  opts = _parse_opts_from_request( request )

  # If a fixture is used by a test class, this seems to be the only
  # way to retrieve the fixture value.
  # https://stackoverflow.com/a/37761165/13190001
  if request.cls is not None:
    request.cls.cmdline_opts = opts

  return opts

@pytest.fixture
def no_translation( request ):
  """Mark a test case as not to be translated."""
  opts = _parse_opts_from_request( request )
  if opts['test_verilog'] != '':
    pytest.skip("skipping untranslatable test cases with --test-verilog")

def pytest_configure(config):
  pass

def pytest_unconfigure(config):
  pass

def pytest_cmdline_preparse(config, args):
  """Don't write *.pyc and __pycache__ files."""
  import sys
  sys.dont_write_bytecode = True

def pytest_runtest_setup(item):
  if _any_opts_present(item.config) and 'cmdline_opts' not in item.fixturenames:
    pytest.skip("'cmdline_opts' is required by pytest commandline but not used")

#-------------------------------------------------------------------------
# helper functions
#-------------------------------------------------------------------------

def _any_opts_present( config ):
  opt_default_pairs = [
      ( 'test_verilog',  ''   ),
      ( 'dump_textwave', None ),
      ( 'dump_vcd',      None ),
      ( 'dump_vtb',      None ),
      ( 'max_cycles',    None ),
  ]
  return any([config.getoption(opt) != val for opt, val in opt_default_pairs])

def _parse_opts_from_request( request ):
  opts = {}

  # test_verilog
  test_verilog = request.config.getoption("test_verilog")
  try:
    test_verilog = int(test_verilog)
  except ValueError:
    assert test_verilog in ['', 'zeros', 'ones', 'rand'], \
        "--test-verilog should be an int or one of '', 'zeros', 'ones', 'rand'!"
  opts['test_verilog'] = test_verilog

  # dump_textwave
  dump_textwave = request.config.getoption("dump_textwave")
  opts['dump_textwave'] = bool(dump_textwave)

  # dump_vcd
  dump_vcd = request.config.getoption("dump_vcd")
  if dump_vcd:
    test_module = request.module.__name__
    test_name   = request.node.name.replace('-', '_').replace( '[', '_' ).replace( ']', '' )
    dump_vcd = f'{test_module}__{test_name}'
  else:
    dump_vcd = ''
  opts['dump_vcd'] = dump_vcd

  # dump_vtb
  dump_vtb = request.config.getoption("dump_vtb")
  if dump_vtb:
    assert request.config.getoption("test_verilog"), "--dump-vtb requires --test-verilog"
    test_module = request.module.__name__
    test_name   = request.node.name
    dump_vtb    = test_name.replace('-', '_').replace( '[', '_' ).replace( ']', '' )
  else:
    dump_vtb = ''
  opts['dump_vtb'] = dump_vtb

  # max_cycles
  max_cycles = request.config.getoption("max_cycles")
  if max_cycles is not None:
    try:
      max_cycles = int(max_cycles)
    except ValueError:
      raise Exception("command line option `--max-cycles` should have integer value!")
  opts['max_cycles'] = max_cycles

  return opts



#=========================================================================
# BehavioralTranslatorL5.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 22, 2019
"""Provide L5 behavioral translator."""

from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRGenL5Pass import (
    BehavioralRTLIRGenL5Pass,
)
from pymtl3.passes.rtlir.behavioral.BehavioralRTLIRTypeCheckL5Pass import (
    BehavioralRTLIRTypeCheckL5Pass,
)

from .BehavioralTranslatorL4 import BehavioralTranslatorL4


class BehavioralTranslatorL5( BehavioralTranslatorL4 ):
  def __init__( s, top ):
    super().__init__( top )

  def clear( s, tr_top ):
    super().clear( tr_top )

  #-----------------------------------------------------------------------
  # _gen_behavioral_trans_metadata
  #-----------------------------------------------------------------------

  # Override
  def _gen_behavioral_trans_metadata( s, m ):
    m.apply( BehavioralRTLIRGenL5Pass( s.tr_top ) )
    m.apply( BehavioralRTLIRTypeCheckL5Pass( s.tr_top ) )
    s.behavioral.rtlir[m] = \
        m.get_metadata( BehavioralRTLIRGenL5Pass.rtlir_upblks )
    s.behavioral.freevars[m] =\
        m.get_metadata( BehavioralRTLIRTypeCheckL5Pass.rtlir_freevars )
    s.behavioral.tmpvars[m] =\
        m.get_metadata( BehavioralRTLIRTypeCheckL5Pass.rtlir_tmpvars )

    # Visit the whole component hierarchy because now we have subcomponents
    for child in m.get_child_components(repr):
      s._gen_behavioral_trans_metadata( child )

  #-----------------------------------------------------------------------
  # translate_behavioral
  #-----------------------------------------------------------------------

  # Override
  def translate_behavioral( s, m ):
    super().translate_behavioral( m )
    for child in m.get_child_components(repr):
      s.translate_behavioral( child )



#=========================================================================
# VBehavioralTranslatorL5.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 18, 2019
"""Provide the level 5 SystemVerilog translator implementation."""

from pymtl3.passes.backends.generic.behavioral.BehavioralTranslatorL5 import (
    BehavioralTranslatorL5,
)
from pymtl3.passes.rtlir import BehavioralRTLIR as bir
from pymtl3.passes.rtlir import RTLIRType as rt

from ...errors import VerilogTranslationError
from .VBehavioralTranslatorL4 import (
    BehavioralRTLIRToVVisitorL4,
    VBehavioralTranslatorL4,
)


class VBehavioralTranslatorL5(
    VBehavioralTranslatorL4, BehavioralTranslatorL5):

  def _get_rtlir2v_visitor( s ):
    return BehavioralRTLIRToVVisitorL5

#-------------------------------------------------------------------------
# BehavioralRTLIRToVVisitorL5
#-------------------------------------------------------------------------

class BehavioralRTLIRToVVisitorL5( BehavioralRTLIRToVVisitorL4 ):
  """Visitor that translates RTLIR to SystemVerilog for a single upblk."""

  #-----------------------------------------------------------------------
  # visit_Attribute
  #-----------------------------------------------------------------------

  def visit_Attribute( s, node ):
    """Return the SystemVerilog representation of an attribute.

    Add support for subcomponent attributes in L5.
    """
    # Generate subcomponent attribute
    if isinstance( node.value.Type, rt.Component ) and \
       not isinstance( node.value, bir.Base ):

      value = s.visit( node.value )
      attr = node.attr
      s.check_res( node, attr )
      return s.process_unpacked_q( node,
                f'{value}__{attr}',
                f'{value}__{attr}{{}}' )

    return super().visit_Attribute( node )

  #-----------------------------------------------------------------------
  # visit_Index
  #-----------------------------------------------------------------------

  def visit_Index( s, node ):
    if isinstance( node.value.Type, rt.Array ) and \
       isinstance( node.value.Type.get_sub_type(), rt.Component ):
      idx = s.visit( node.idx )
      s._unpacked_q.appendleft(idx)
      value = s.visit( node.value )
      return value

    else:
      return super().visit_Index( node )



#=========================================================================
# VBehavioralTranslatorL2_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 28, 2019
"""Test the SystemVerilog translator implementation."""

import pytest

from pymtl3 import *
from pymtl3.passes.backends.verilog.util.utility import verilog_reserved
from pymtl3.passes.rtlir import BehavioralRTLIRGenPass, BehavioralRTLIRTypeCheckPass

from ....testcases import (
    CaseBoolTmpVarComp,
    CaseElifBranchComp,
    CaseFixedSizeSliceComp,
    CaseForLoopEmptySequenceComp,
    CaseForRangeLowerUpperStepPassThroughComp,
    CaseIfBasicComp,
    CaseIfBoolOpInForStmtComp,
    CaseIfDanglingElseInnerComp,
    CaseIfDanglingElseOutterComp,
    CaseIfExpBothImplicitComp,
    CaseIfExpInForStmtComp,
    CaseIfExpUnaryOpInForStmtComp,
    CaseIfTmpVarInForStmtComp,
    CaseLambdaConnectComp,
    CaseLambdaConnectWithListComp,
    CaseNestedIfComp,
    CaseReducesInx3OutComp,
    CaseTmpVarInUpdateffComp,
)
from ..VBehavioralTranslatorL2 import BehavioralRTLIRToVVisitorL2


def run_test( case, m ):
  m.elaborate()
  m.apply( BehavioralRTLIRGenPass( m ) )
  m.apply( BehavioralRTLIRTypeCheckPass( m ) )

  visitor = BehavioralRTLIRToVVisitorL2(lambda x: x in verilog_reserved)
  upblks = m.get_metadata( BehavioralRTLIRGenPass.rtlir_upblks )
  m_all_upblks = m.get_update_blocks()
  assert len(m_all_upblks) == 1

  for blk in m_all_upblks:
    upblk_src = visitor.enter( blk, upblks[blk] )
    upblk_src = "\n".join( upblk_src )
    assert upblk_src + '\n' == case.REF_UPBLK

@pytest.mark.parametrize(
    'case', [
      CaseReducesInx3OutComp,
      CaseIfBasicComp,
      CaseIfDanglingElseInnerComp,
      CaseIfDanglingElseOutterComp,
      CaseElifBranchComp,
      CaseNestedIfComp,
      CaseForLoopEmptySequenceComp,
      CaseForRangeLowerUpperStepPassThroughComp,
      CaseIfExpInForStmtComp,
      CaseIfExpBothImplicitComp,
      CaseIfBoolOpInForStmtComp,
      CaseIfTmpVarInForStmtComp,
      CaseFixedSizeSliceComp,
      CaseLambdaConnectComp,
      CaseLambdaConnectWithListComp,
      CaseBoolTmpVarComp,
      CaseTmpVarInUpdateffComp,
    ]
)
def test_verilog_behavioral_L2( case ):
  run_test( case, case.DUT() )



#=========================================================================
# BehavioralRTLIRGenL1Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : Oct 20, 2018
"""Provide L1 behavioral RTLIR generation pass."""
import ast
import copy
import sys

from pymtl3 import MetadataKey, dsl
from pymtl3.datatypes import (
    Bits,
    concat,
    is_bitstruct_class,
    is_bitstruct_inst,
    reduce_and,
    reduce_or,
    reduce_xor,
    sext,
    trunc,
    zext,
)
from pymtl3.passes.rtlir.errors import PyMTLSyntaxError
from pymtl3.passes.rtlir.RTLIRPass import RTLIRPass
from pymtl3.passes.rtlir.rtype.RTLIRType import RTLIRGetter
from pymtl3.passes.rtlir.util.utility import get_ordered_upblks, get_ordered_update_ff

from . import BehavioralRTLIR as bir


class BehavioralRTLIRGenL1Pass( RTLIRPass ):

  # Pass metadata

  #: A dictionary that maps upblk functions to their BIR representation
  #:
  #: Type: ``dict``; output
  rtlir_upblks = MetadataKey()

  def __init__( s, translation_top ):
    c = s.__class__
    s.tr_top = translation_top
    if not translation_top.has_metadata( c.rtlir_getter ):
      translation_top.set_metadata( c.rtlir_getter, RTLIRGetter(cache=True) )

  def __call__( s, m ):
    """Generate RTLIR for all upblks of m."""
    c = s.__class__

    if m.has_metadata( c.rtlir_upblks ):
      rtlir_upblks = m.get_metadata( c.rtlir_upblks )
    else:
      rtlir_upblks = {}
      m.set_metadata( c.rtlir_upblks, rtlir_upblks )

    visitor = s.get_rtlir_generator_class()( m )
    upblks = {
      bir.CombUpblk : get_ordered_upblks(m),
      bir.SeqUpblk  : get_ordered_update_ff(m),
    }
    # Sort the upblks by their name
    upblks[bir.CombUpblk].sort( key = lambda x: x.__name__ )
    upblks[bir.SeqUpblk ].sort( key = lambda x: x.__name__ )

    for upblk_type in ( bir.CombUpblk, bir.SeqUpblk ):
      for blk in upblks[ upblk_type ]:
        visitor._upblk_type = upblk_type
        upblk_info = m.get_update_block_info( blk )
        upblk = visitor.enter( blk, upblk_info[-1] )
        upblk.is_lambda = upblk_info[0]
        upblk.src       = upblk_info[1]
        upblk.lino      = upblk_info[2]
        upblk.filename  = upblk_info[3]
        rtlir_upblks[ blk ] = upblk

  def get_rtlir_generator_class( s ):
    return BehavioralRTLIRGeneratorL1

class BehavioralRTLIRGeneratorL1( ast.NodeVisitor ):
  def __init__( s, component ):
    s.component = component

    if sys.version_info < (3,8,10000):
      s.visit_Subscript = s._visit_Subscript_up_to_py38
    else:
      s.visit_Subscript = s._visit_Subscript_starting_py39

  def enter( s, blk, ast ):
    """Entry point of RTLIR generation."""
    s.blk     = blk

    # s.globals contains a dict of the global namespace of the module where
    # blk was defined
    s.globals = blk.__globals__

    # s.closure contains the free variables defined in an enclosing scope.
    # Basically this is the model instance s.
    s.closure = {}

    for i, var in enumerate( blk.__code__.co_freevars ):
      try:
        s.closure[ var ] = blk.__closure__[ i ].cell_contents
      except ValueError:
        pass

    s.const_extractor = ConstantExtractor( s.blk, s.globals, s.closure )
    ret = s.visit( ast )
    ret.component = s.component
    return ret

  def handle_constant( s, node, obj ):
    if isinstance( obj, int ):
      return bir.Number( obj )
    elif isinstance( obj, Bits ):
      return bir.SizeCast( obj.nbits, bir.Number( obj.uint() ) )
    else:
      return None

  def get_call_obj( s, node ):
    if hasattr(node, "starargs") and node.starargs:
      raise PyMTLSyntaxError( s.blk, node, 'star argument is not supported!')
    if hasattr(node, "kwargs") and node.kwargs:
      raise PyMTLSyntaxError( s.blk, node,
        'double-star argument is not supported!')
    if node.keywords:
      raise PyMTLSyntaxError( s.blk, node, 'keyword argument is not supported!')

    obj = s.const_extractor.enter( node.func )
    if obj is not None:
      return obj
    else:
      raise PyMTLSyntaxError( s.blk, node, f'{node.func} function is not found!' )

  def visit_Module( s, node ):
    if len( node.body ) != 1 or \
        not isinstance( node.body[0], ast.FunctionDef ):
      raise PyMTLSyntaxError( s.blk, node,
        'Update blocks should have exactly one FuncDef!' )
    ret = s.visit( node.body[0] )
    ret.ast = node
    return ret

  def visit_FunctionDef( s, node ):
    """Return the behavioral RTLIR of function node.

    We do not need to check the decorator list -- the fact that we are
    visiting this node ensures this node was added to the upblk
    dictionary through update() (or other PyMTL decorators) earlier!
    """
    # Check the arguments of the function
    if node.args.args or node.args.vararg or node.args.kwarg:
      raise PyMTLSyntaxError( s.blk, node,
        'Update blocks should not have arguments!' )

    # Save the name of the upblk
    s._upblk_name = node.name

    # Construct the node using the type of upblk
    ret = s._upblk_type( node.name, [] )

    for stmt in node.body:
      ret.body.append( s.visit( stmt ) )
    ret.ast = node
    return ret

  def visit_Assign( s, node ):

    if len( node.targets ) < 1:
      raise PyMTLSyntaxError( s.blk, node,
        'At least one assignment target should be provided!' )

    value = s.visit( node.value )
    targets = [ s.visit( target ) for target in node.targets ]
    ret = bir.Assign( targets, value, False ) # Need a handle to bir node

    # Determine if this is a blocking/non-blocking assignment
    ret.blocking = s.get_blocking(node, ret)

    ret.ast = node
    return ret

  def get_blocking( s, node, bir_node ):
    return s._upblk_type is bir.CombUpblk

  def visit_AugAssign( s, node ):
    """Return the behavioral RTLIR of a non-blocking assignment

    If the given AugAssign is not @= or <<=, throw PyMTLSyntaxError
    """
    if isinstance( node.op, (ast.LShift, ast.MatMult) ):
      value = s.visit( node.value )
      targets = [ s.visit( node.target ) ]
      blocking = False if isinstance(node.op, ast.LShift) else True
      ret = bir.Assign( targets, value, blocking )
      ret.ast = node
      return ret
    raise PyMTLSyntaxError( s.blk, node,
        'invalid operation: augmented assignment is not @= or <<= assignment!' )

  def visit_Call( s, node ):
    """Return the behavioral RTLIR of method calls.

    Some data types are interpreted as function calls in the Python AST.
    Example: Bits4(2)
    These are converted to different RTLIR nodes in different contexts.
    """
    num_args = len( node.args )

    obj = s.get_call_obj( node )
    if ( obj == copy.copy ) or ( obj == copy.deepcopy ):
      if num_args != 1:
        raise PyMTLSyntaxError( s.blk, node,
          f'copy method {obj} takes exactly 1 argument!')
      ret = s.visit( node.args[0] )
      ret.ast = node
      return ret

    # Now that we have the live Python object, there are a few cases that
    # we need to treat separately:
    # 1. Instantiation: Bits16( 10 ) where obj is an instance of Bits
    # Bits16( 1+2 ), Bits16( s.STATE_A )?
    # 2. concat()
    # 3. zext(), sext()
    # TODO: support the following
    # 4. reduce_and(), reduce_or(), reduce_xor()
    # 5. Real function call: not supported yet

    # Deal with Bits type cast
    if isinstance(obj, type) and issubclass( obj, Bits ):
      nbits = obj.nbits
      if num_args > 1:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly one or zero argument should be given to Bits!' )
      if num_args == 0:
        ret = bir.SizeCast( nbits, bir.Number( 0 ) )
      else:
        ret = bir.SizeCast( nbits, s.visit( node.args[0] ) )

    # concat method
    elif obj is concat:
      if num_args < 1:
        raise PyMTLSyntaxError( s.blk, node,
          'at least one argument should be given to concat!' )
      values = [s.visit(c) for c in node.args]
      ret = bir.Concat( values )

    # zext method
    elif obj is zext:
      if num_args != 2:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly two arguments should be given to zext!' )

      nbits = s.const_extractor.enter( node.args[1] )
      if isinstance(nbits, type) and issubclass( nbits, Bits ):
        nbits = nbits.nbits
      if not isinstance( nbits, int ):
        raise PyMTLSyntaxError( s.blk, node,
          'the 2nd argument of zext {nbits} is not a constant int or BitsN type!' )

      ret = bir.ZeroExt( nbits, s.visit( node.args[0] ) )

    # sext method
    elif obj is sext:
      if num_args != 2:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly two arguments should be given to sext!' )

      nbits = s.const_extractor.enter( node.args[1] )
      if isinstance(nbits, type) and issubclass( nbits, Bits ):
        nbits = nbits.nbits
      if not isinstance( nbits, int ):
        raise PyMTLSyntaxError( s.blk, node,
          'the 2nd argument of sext {nbits} is not a constant int or BitsN type!' )

      ret = bir.SignExt( nbits, s.visit( node.args[0] ) )

    # trunc method
    elif obj is trunc:
      if num_args != 2:
        raise PyMTLSyntaxError( s.blk, node,
          'exactly two arguments should be given to trunc!' )

      nbits = s.const_extractor.enter( node.args[1] )
      if isinstance(nbits, type) and issubclass( nbits, Bits ):
        nbits = nbits.nbits
      if not isinstance( nbits, int ):
        raise PyMTLSyntaxError( s.blk, node,
          'the 2nd argument of trunc {nbits} is not a constant int or BitsN type!' )

      ret = bir.Truncate( nbits, s.visit( node.args[0] ) )

    # reduce methods
    elif obj is reduce_and or obj is reduce_or or obj is reduce_xor:
      if obj is reduce_and:
        op = bir.BitAnd()
      elif obj is reduce_or:
        op = bir.BitOr()
      elif obj is reduce_xor:
        op = bir.BitXor()
      if num_args != 1:
        raise PyMTLSyntaxError( s.blk, node,
          f'exactly two arguments should be given to reduce {op} methods!' )

      ret = bir.Reduce( op, s.visit( node.args[0] ) )

    else:
      # Only Bits class instantiation is supported at L1
      raise PyMTLSyntaxError( s.blk, node, f'Unrecognized method call {obj.__name__}!' )

    ret.ast = node
    return ret

  def visit_Attribute( s, node ):
    obj = s.const_extractor.enter( node )
    ret = s.handle_constant( node, obj )
    if ret:
      return ret

    ret = bir.Attribute( s.visit( node.value ), node.attr )
    ret.ast = node
    return ret

  def _visit_Subscript_up_to_py38( s, node ):
    obj = s.const_extractor.enter( node )
    ret = s.handle_constant( node, obj )
    if ret:
      return ret

    value = s.visit( node.value )
    if isinstance( node.slice, ast.Slice ):
      if node.slice.step is not None:
        raise PyMTLSyntaxError( s.blk, node,
          'Slice with steps is not supported!' )
      lower, upper = s.visit( node.slice )
      ret = bir.Slice( value, lower, upper )
      ret.ast = node
      return ret

    # signal[ index ]
    # index might be a slice object!
    if isinstance( node.slice, ast.Index ):
      idx = s.visit( node.slice )
      # If we have a static slice object then use it
      if isinstance( idx, bir.FreeVar ) and isinstance( idx.obj, slice ):
        slice_obj = idx.obj
        if slice_obj.step is not None:
          raise PyMTLSyntaxError( s.blk, node,
            'Slice with steps is not supported!' )
        assert isinstance( slice_obj.start, int ) and \
               isinstance( slice_obj.stop, int ), \
            f"start and stop of slice object {slice_obj} must be integers!"
        ret = bir.Slice( value,
              bir.Number(slice_obj.start), bir.Number(slice_obj.stop) )
      # Else this is a real index
      else:
        ret = bir.Index( value, idx )
      ret.ast = node
      return ret

    raise PyMTLSyntaxError( s.blk, node,
      'Illegal subscript ' + node + ' encountered!' )

  def _visit_Subscript_starting_py39( s, node ):
    obj = s.const_extractor.enter( node )
    ret = s.handle_constant( node, obj )
    if ret:
      return ret

    value = s.visit( node.value )
    if isinstance( node.slice, ast.Slice ):
      if node.slice.step is not None:
        raise PyMTLSyntaxError( s.blk, node,
          'Slice with steps is not supported!' )
      lower, upper = s.visit( node.slice )
      ret = bir.Slice( value, lower, upper )
      ret.ast = node
      return ret

    # signal[ index ]
    # index might be a slice object!
    idx = s.visit( node.slice )
    # If we have a static slice object then use it
    if isinstance( idx, bir.FreeVar ) and isinstance( idx.obj, slice ):
      slice_obj = idx.obj
      if slice_obj.step is not None:
        raise PyMTLSyntaxError( s.blk, node,
          'Slice with steps is not supported!' )
      assert isinstance( slice_obj.start, int ) and \
             isinstance( slice_obj.stop, int ), \
          f"start and stop of slice object {slice_obj} must be integers!"
      ret = bir.Slice( value,
            bir.Number(slice_obj.start), bir.Number(slice_obj.stop) )
    # Else this is a real index
    else:
      ret = bir.Index( value, idx )
    ret.ast = node
    return ret

  def visit_Slice( s, node ):
    return ( s.visit( node.lower ), s.visit( node.upper ) )

  def visit_Index( s, node ):
    return s.visit( node.value )

  def visit_Name( s, node ):
    if node.id in s.closure:
      # free var from closure
      obj = s.closure[ node.id ]
      if isinstance( obj, dsl.Component ):
        # Component freevars are an L1 thing.
        if obj is not s.component:
          raise PyMTLSyntaxError( s.blk, node,
            f'Component {obj} is not a sub-component of {s.component}!' )
        ret = bir.Base( obj )
      else:
        # A closure variable could be a loop index. We need to
        # generate per-function closure variable instead of assuming
        # they will have the same value.
        ret = bir.FreeVar( f"{node.id}_at_{s.blk.__name__}", obj )
      ret.ast = node
      return ret
    elif node.id in s.globals:
      # free var from the global name space
      # For now we can still safely assume all upblks will see the same
      # value for a free var from the global space?
      ret = bir.FreeVar( node.id, s.globals[ node.id ] )
      ret.ast = node
      return ret
    raise PyMTLSyntaxError( s.blk, node,
      f'Temporary variable {node.id} is not supported at L1!' )

  def visit_Num( s, node ):
    ret = bir.Number( node.n )
    ret.ast = node
    return ret

  def visit_If( s, node ): raise NotImplementedError()

  def visit_For( s, node ): raise NotImplementedError()

  def visit_BoolOp( s, node ): raise NotImplementedError()

  def visit_BinOp( s, node ): raise NotImplementedError()

  def visit_UnaryOp( s, node ): raise NotImplementedError()

  def visit_IfExp( s, node ): raise NotImplementedError()

  def visit_Compare( s, node ): raise NotImplementedError()

  # $display
  def visit_Print( s, node ): raise NotImplementedError()

  # function
  def visit_Return( s, node ): raise NotImplementedError()

  # SV assertion
  def visit_Assert( s, node ): raise NotImplementedError()

  def visit_Expr( s, node ):
    """Return the behavioral RTLIR of an expression.

    ast.Expr might be useful when a statement is only a call to a task or
    a non-returning function.
    """
    raise PyMTLSyntaxError(
      s.blk, node, 'Stand-alone expression is not supported yet!' )

  def visit_Lambda( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: lambda function' )

  def visit_Dict( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: dict' )

  def visit_Set( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: set' )

  def visit_List( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: list' )

  def visit_Tuple( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid type: tuple' )

  def visit_ListComp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: list comprehension' )

  def visit_SetComp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: set comprehension' )

  def visit_DictComp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: dict comprehension' )

  def visit_GeneratorExp( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: generator expression' )

  def visit_Yield( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: yield' )

  def visit_Repr( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: repr' )

  def visit_Str( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: str' )

  def visit_ClassDef( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: classdef' )

  def visit_Delete( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: delete' )

  def visit_With( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: with' )

  def visit_Raise( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: raise' )

  def visit_TryExcept( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: try-except' )

  def visit_TryFinally( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: try-finally' )

  def visit_Import( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: import' )

  def visit_ImportFrom( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: import-from' )

  def visit_Exec( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: exec' )

  def visit_Global( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: global' )

  def visit_Pass( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: pass' )

  def visit_Break( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: break' )

  def visit_Continue( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: continue' )

  def visit_While( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: while' )

  def visit_ExtSlice( s, node ):
    raise PyMTLSyntaxError( s.blk, node, 'invalid operation: extslice' )

class ConstantExtractor( ast.NodeVisitor ):
  def __init__( s, blk, global_ns, closure_ns ):
    s.blk = blk
    s.globals = global_ns
    s.cache = {}
    s.closure = closure_ns
    s.pymtl_functions = { concat, sext, zext, trunc,
                          reduce_or, reduce_and, reduce_xor,
                          copy.copy, copy.deepcopy }

    if sys.version_info < (3,8,10000):
      s.visit_Subscript = s._visit_Subscript_up_to_py38
    else:
      s.visit_Subscript = s._visit_Subscript_starting_py39

  def generic_visit( s, node ):
    return None

  def enter( s, node ):
    ret = s.visit( node )
    # Constant objects that are recognized
    # 1. int, BitsN( X )
    # 2. BitsN
    # 3. BitStruct, BitStruct()
    # 4. Functions, including concat, zext, sext, etc.
    is_value = isinstance(ret, (int, Bits)) or is_bitstruct_inst(ret)
    is_type = isinstance(ret, type) and (issubclass(ret, Bits) or is_bitstruct_class(ret))
    try:
      is_function = ret in s.pymtl_functions
    except:
      is_function = False

    if is_value or is_type or is_function:
      return ret
    else:
      return None

  def visit_Attribute( s, node ):
    if node in s.cache:
      return s.cache[node]
    value = s.visit( node.value )
    try:
      ret = getattr( value, node.attr )
    except AttributeError:
      ret = None
    s.cache[node] = ret
    return ret

  def _visit_Subscript_up_to_py38( s, node ):
    if node in s.cache:
      return s.cache[node]
    ret = None
    if isinstance( node.slice, ast.Index ):
      value = s.visit( node.value )
      idx = s.visit( node.slice )
      if value is not None and idx is not None:
        try:
          ret = value[idx]
        except:
          ret = None
    s.cache[node] = ret
    return ret

  def _visit_Subscript_starting_py39( s, node ):
    if node in s.cache:
      return s.cache[node]
    ret = None
    value = s.visit( node.value )
    idx = s.visit( node.slice )
    if value is not None and idx is not None:
      try:
        ret = value[idx]
      except:
        ret = None
    s.cache[node] = ret
    return ret

  def visit_Index( s, node ):
    if node in s.cache:
      return s.cache[node]
    ret = s.visit( node.value )
    s.cache[node] = ret
    return ret

  def visit_Name( s, node ):
    name = node.id
    if name in s.closure:
      # free var from closure
      obj = s.closure[ name ]
    elif name in s.globals:
      # free var from the global name space
      obj = s.globals[ name ]
    else:
      obj = None
    return obj

  def visit_Num( s, node ):
    return node.n



#=========================================================================
# VTranslator.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 15, 2019
"""Provide SystemVerilog translator."""

from collections import deque

from pymtl3.passes.backends.generic import RTLIRTranslator
from pymtl3.passes.backends.verilog.errors import VerilogStructuralTranslationError
from pymtl3.passes.backends.verilog.util.utility import verilog_reserved

from ..VerilogPlaceholderPass import VerilogPlaceholderPass
from .behavioral import VBehavioralTranslator as V_BTranslator
from .structural import VStructuralTranslator as V_STranslator


def mk_VTranslator( _RTLIRTranslator, _STranslator, _BTranslator ):

  class _VTranslator( _RTLIRTranslator, _STranslator, _BTranslator ):

    def get_pretty( s, namespace, attr, newline=True ):
      ret = getattr(namespace, attr, "")
      if newline and (ret and ret[-1] != '\n'):
        ret += "\n"
      return ret

    def is_verilog_reserved( s, name ):
      return name in verilog_reserved

    def set_header( s ):
      s.header_keywords = 'generated by PyMTL SystemVerilog translation pass'
      s.header = \
"""\
//-------------------------------------------------------------------------
// {name}.v
//-------------------------------------------------------------------------
// This file is generated by PyMTL SystemVerilog translation pass.

"""

    def rtlir_tr_initialize( s ):
      # Unpacked array indice that will be pushed to the end of signal expr
      s._rtlir_tr_unpacked_q = deque()
      s._placeholder_pass = VerilogPlaceholderPass
      s._mangled_placeholder_top_module_name = ''
      s._included_pickled_files = set()

    def rtlir_tr_src_layout( s, hierarchy ):
      # Sanity check on BitStructs
      all_struct_names = { x.cls.__name__ for x in hierarchy.decl_type_struct }

      for struct in hierarchy.decl_type_struct.keys():
        for field_name in struct.get_all_properties().keys():
          if field_name in all_struct_names:
            raise VerilogStructuralTranslationError(struct,
              f'field {field_name} has the same name as BitStruct type {field_name}!')

      s.set_header()
      name = s._top_module_full_name

      if s.header_keywords not in hierarchy.component_src:
        ret = s.header.format( **locals() )
      else:
        # This is a previous translation result
        ret = ''

      # Add struct definitions
      for struct_dtype, tplt in hierarchy.decl_type_struct.items():
        template = \
"""\
// PyMTL BitStruct {dtype_name} Definition
{struct_def}\
"""
        dtype_name = struct_dtype.get_name()
        struct_def = tplt['def'] + '\n'
        ret += template.format( **locals() )

      # Add component sources
      ret += hierarchy.component_src
      return ret

    def rtlir_tr_components( s, components ):
      return "\n\n".join( components.values() )

    def rtlir_tr_component( s, behavioral, structural ):
      component_name = structural.component_name
      file_info = structural.component_file_info
      full_name = structural.component_full_name

      if structural.component_explicit_module_name:
        module_name = structural.component_explicit_module_name
      elif structural.component_is_top and s._mangled_placeholder_top_module_name:
        module_name = s._mangled_placeholder_top_module_name
      else:
        module_name = structural.component_unique_name

      if structural.component_no_synthesis:
        no_synth_begin = '`ifndef SYNTHESIS\n'
        no_synth_end   = '`endif'
      else:
        no_synth_begin = ''
        no_synth_end   = ''

      s._top_module_name = structural.component_name
      s._top_module_full_name = module_name

      if full_name != module_name:
        optional_full_name = f"Full name: {full_name}\n// "
      else:
        optional_full_name = ""

      if structural.placeholder_src:
        # This is a placeholder
        placeholder_src = structural.placeholder_src

        if '// PyMTL VerilogPlaceholder' not in structural.placeholder_src:
          template = \
"""\
// PyMTL VerilogPlaceholder {component_name} Definition
// {optional_full_name}At {file_info}

{no_synth_begin}{placeholder_src}
{no_synth_end}"""
        else:
          # This is a previous translation result
          template = '{placeholder_src}'

        return template.format( **locals() )

      else:
        template = \
"""\
// PyMTL Component {component_name} Definition
// {optional_full_name}At {file_info}

{no_synth_begin}module {module_name}
(
{ports});
{body}
endmodule
{no_synth_end}"""
        ports_template = "{port_decls}{ifc_decls}"

        port_decls = s.get_pretty(structural, 'decl_ports', False)
        ifc_decls = s.get_pretty(structural, 'decl_ifcs', False)
        if port_decls or ifc_decls:
          if port_decls and ifc_decls:
            port_decls += ',\n'
          ifc_decls += '\n'
        ports = ports_template.format(**locals())

        const_decls = s.get_pretty(structural, "decl_consts")
        fvar_decls = s.get_pretty(behavioral, "decl_freevars")
        wire_decls = s.get_pretty(structural, "decl_wires")
        tmpvar_decls = s.get_pretty(behavioral, "decl_tmpvars")
        subcomp_decls = s.get_pretty(structural, "decl_subcomps")
        upblk_decls = s.get_pretty(behavioral, "upblk_decls")
        body = const_decls + fvar_decls + wire_decls + subcomp_decls \
             + tmpvar_decls + upblk_decls
        connections = s.get_pretty(structural, "connections")
        if (body and connections) or (not body and connections):
          connections = '\n' + connections
        body += connections

        return template.format( **locals() )

  return _VTranslator

VTranslator = mk_VTranslator( RTLIRTranslator, V_STranslator, V_BTranslator )



"""
========================================================================
XcelMsg.py
========================================================================
Accelerator message type implementation.

Author : Yanghui Ou
Date   : June 3, 2019
"""
from pymtl3 import *


class XcelMsgType:
  # TODO: figure out whether we want to use Bits1 here.
  READ  = 0
  WRITE = 1

  str = {
    READ  : "rd",
    WRITE : "wr",
  }

def mk_xcel_msg( addr, data ):
  return mk_xcel_req_msg( addr, data ), mk_xcel_resp_msg( data )

def mk_xcel_req_msg( a, d ):
  @bitstruct
  class XcelReqMsg:
    type_ : Bits1
    addr  : mk_bits( a )
    data  : mk_bits( d )

    def __str__( self ):
      return "{}:{}:{}".format(
        XcelMsgType.str[ int(self.type_) ],
        self.addr,
        self.data if self.type_ != XcelMsgType.READ else " " * ( d//4 ),
      )

  return XcelReqMsg

def mk_xcel_resp_msg( d ):
  @bitstruct
  class XcelRespMsg:
    type_ : Bits1
    data  : mk_bits( d )

    def __str__( self ):
      return "{}:{}".format(
        XcelMsgType.str[ int(self.type_) ],
        self.data if self.type_ != XcelMsgType.WRITE else " " * ( d//4 ),
      )

  return XcelRespMsg



#=========================================================================
# VerilogVerilatorImportConfigs.py
#=========================================================================
# Author : Peitian Pan
# Date   : Jul 28, 2019
"""Configuration of Verilator import pass."""

import copy
import os
import subprocess
from textwrap import fill, indent

from pymtl3.passes.errors import InvalidPassOptionValue
from pymtl3.passes.PassConfigs import BasePassConfigs, Checker
from pymtl3.passes.PlaceholderConfigs import expand

from ..util.utility import get_hash_of_lean_verilog
from .VerilogVerilatorImportPass import VerilogVerilatorImportPass


class VerilogVerilatorImportConfigs( BasePassConfigs ):

  Options = {
    # Import this component?
    "enable" : False,

    # Enable verbose mode?
    "verbose" : False,

    # Trade off compilation time for fast simulation performance?
    # controls: vl_opt_level, vl_unroll_count
    "fast": False,

    # Verilator optimization options

    # Enable external line trace?
    # Once enabled, the `line_trace()` method of the imported component
    # will return a string read from the external `line_trace()` function.
    # This means your Verilog module has to have a `line_trace` function
    # that provides the line trace string which has less than 512 characters.
    # Default to False
    "vl_line_trace" : False,

    # Enable all verilator coverage
    "vl_coverage" : False,

    # Enable all verilator coverage
    "vl_line_coverage" : False,

    # Enable all verilator coverage
    "vl_toggle_coverage" : False,

    # Verilator code generation options
    # These options will be passed to verilator to generate the C simulator.
    # By default, verilator is called with `--cc`.

    # --Mdir
    # Expects the path of Makefile output directory;
    # "" to use `obj_dir_<translated_top_module>`
    "vl_mk_dir" : "",

    # --assert
    # Expects a boolean value
    "vl_enable_assert" : True,

    # Verilator warning-related options

    # False to disable the warnings, True to enable
    "vl_W_lint" : True,
    "vl_W_style" : True,
    "vl_W_fatal" : True,

    # Un-warn all warnings in the given list; [] to disable this option
    # The given list should only include strings that appear in `Warnings`
    "vl_Wno_list" : [ 'UNOPTFLAT', 'UNSIGNED', 'WIDTH' ],

    # Verilator misc options

    # What is the inital value of signals?
    # Should be one of ['zeros', 'ones', 'rand']
    "vl_xinit" : "zeros",

    # --trace
    # Expects a boolean value
    "vl_trace" : False,

    # The output filename of Verilator VCD tracing
    # default is {component_name}.verilator1
    "vl_trace_filename" : "",

    # Passed to verilator tracing function
    "vl_trace_timescale" : "10ps",

    # `vl_trace_cycle_time`*`vl_trace_timescale` is the cycle time of the
    # PyMTL clock that appears in the generated VCD
    # With the default options, the frequency of PyMTL clock is 1GHz
    "vl_trace_cycle_time" : 100,

    # Set to true to allow for on-demand VCD dumping
    "vl_trace_on_demand" : False,

    # Top level port name that is used to enable VCD dumping when
    # `vl_trace_on_demand` is True. Assuming the port is an active-high
    # enable signal.
    "vl_trace_on_demand_portname" : "",

    # C-compilation options
    # These options will be passed to the C compiler to create a shared lib.

    # Additional flags to be passed to the C compiler.
    # By default, CC is called with `-O0 -fPIC -shared`.
    # "" to disable this option
    "c_flags" : "",

    # Additional include search path of the C compiler.
    # [] to disable this option
    "c_include_path" : [],

    # Additional C source files passed to the C compiler.
    # [] to compile verilator generated files only.
    "c_srcs" : [],

    # `LDLIBS` will be listed after the primary target file whereas
    # `LDFLAGS` will be listed before.

    # We enforce the GNU makefile implicit rule that `LDFLAGS` should only
    # include non-library linker flags such as `-L`.
    "ld_flags" : "",

    # We enforce the GNU makefile implicit rule that `LDLIBS` should only
    # include library linker flags/names such as `-lfoo`.
    "ld_libs" : "",

    "c_flags" : "",
  }

  Checkers = {
    ("enable", "verbose", "vl_enable_assert", "vl_line_trace", "vl_W_lint", "vl_W_style",
     "vl_W_fatal", "vl_trace", "vl_coverage", "vl_line_coverage", "vl_toggle_coverage",
     "vl_trace_on_demand"):
      Checker( lambda v: isinstance(v, bool), "expects a boolean" ),

    ("c_flags", "ld_flags", "ld_libs", "vl_trace_filename", "vl_trace_on_demand_portname"):
      Checker( lambda v: isinstance(v, str),  "expects a string" ),

    "vl_Wno_list": Checker( lambda v: isinstance(v, list) and all(w in VerilogPlaceholderConfigs.Warnings for w in v),
                            "expects a list of warnings" ),

    "vl_xinit": Checker( lambda v: (v in ['zeros', 'ones', 'rand']) or isinstance(v, int),
                  "vl_xinit should be an integer or one of ['zeros', 'ones', 'rand']" ),

    "vl_trace_timescale": Checker( lambda v: isinstance(v, str) and len(v) > 2 and v[-1] == 's' and \
                                    v[-2] in ['p', 'n', 'u', 'm'] and \
                                    all(c.isdigit() for c in v[:-2]),
                                    "expects a timescale string" ),

    "vl_trace_cycle_time": Checker( lambda v: isinstance(v, int) and (v % 2) == 0,
                                    "expects an integer `n` such that `n`*`vl_trace_timescale` is the cycle time" ),

    "vl_mk_dir": Checker( lambda v: isinstance(v, str), "expects a path to directory" ),

    "c_include_path": Checker( lambda v: isinstance(v, list) and all(os.path.isdir(expand(p)) for p in v),
                                "expects a list of paths to directories" ),

    "c_srcs": Checker( lambda v: isinstance(v, list) and all(os.path.isfile(expand(p)) for p in v),
                       "expects a list of paths to files" )
  }

  Warnings = [
    'ALWCOMBORDER', 'ASSIGNIN', 'ASSIGNDLY', 'BLKANDNBLK', 'BLKSEQ',
    'BLKLOOPINIT', 'BSSPACE', 'CASEINCOMPLETE', 'CASEOVERLAP',
    'CASEX', 'CASEWITHX', 'CDCRSTLOGIC', 'CLKDATA', 'CMPCONST',
    'COLONPLUS', 'COMBDLY', 'CONTASSREG', 'DECLFILENAME', 'DEFPARAM',
    'DETECTARRAY', 'ENDLABEL', 'GENCLK', 'IFDEPTH', 'IGNOREDRETURN',
    'IMPERFECTSCH', 'IMPLICIT', 'IMPORTSTAR', 'IMPURE', 'INCABSPATH',
    'INFINITELOOP', 'INITIALDLY', 'LITENDIAN', 'MODDUP', 'MULTIDRIVEN',
    'MULTITOP', 'PINCONNECTEMPTY', 'PINMISSING', 'PINNOCONNECT',
    'PROCASSWIRE', 'REALCVT', 'REDEFMACRO', 'SELRANGE', 'STMTDLY',
    'SYMRSVDWORD', 'SYNCASYNCNET', 'TASKNSVAR', 'TICKCOUNT', 'UNDRIVEN',
    'UNOPT', 'UNOPTFLAT', 'UNOPTTHREADS', 'UNPACKED', 'UNSIGNED', 'UNUSED',
    'USERINFO', 'USERWARN', 'USERERROR', 'USERFATAL', 'VARHIDDEN', 'WIDTH',
    'WIDTHCONCAT',
  ]

  Pass = VerilogVerilatorImportPass

  #---------------------------------------------------
  # Public APIs
  #---------------------------------------------------

  def setup_configs( s, m, tr_pass, ph_pass ):
    # VerilatorImportConfigs alone does not have the complete information about
    # the module to be imported. For example, we need to read from the placeholder
    # configuration to figure out the pickled file name and the top module name.
    # This method is meant to be called before calling other public APIs.

    s.translated_top_module = m.get_metadata( tr_pass.translated_top_module )
    s.translated_source_file = m.get_metadata( tr_pass.translated_filename )

    ph_cfg = m.get_metadata( ph_pass.placeholder_config )
    s.verilog_hash = get_hash_of_lean_verilog( s.translated_source_file )
    s.v_include = ph_cfg.v_include
    s.v_libs = ph_cfg.v_libs
    s.src_file = ph_cfg.src_file
    s.port_map = ph_cfg.port_map
    s.params = ph_cfg.params

    # Infer vl_line_trace by scanning through the source file
    try:
      trim = lambda s: ''.join(s.split())
      if not s.vl_line_trace:
        with open(s.src_file) as fd:
          for line in fd.readlines():
            if "`VC_TRACE_BEGIN" in trim(line):
              s.vl_line_trace = True
              break
    except OSError:
      pass

    if not s.vl_mk_dir:
      s.vl_mk_dir = f'obj_dir_{s.translated_top_module}'

  def get_vl_xinit_value( s ):
    if s.vl_xinit == 'zeros':
      return 0
    elif s.vl_xinit == 'ones':
      return 1
    elif ( s.vl_xinit == 'rand' ) or isinstance( s.vl_xinit, int ):
      return 2
    else:
      raise InvalidPassOptionValue("vl_xinit should be an int or one of 'zeros', 'ones', or 'rand'!")

  def get_vl_xinit_seed( s ):
    if isinstance( s.vl_xinit, int ):
      return s.vl_xinit
    else:
      return 0

  def get_c_wrapper_path( s ):
    return f'{s.translated_top_module}_v.cpp'

  def get_py_wrapper_path( s ):
    return f'{s.translated_top_module}_v.py'

  def get_shared_lib_path( s ):
    return f'lib{s.translated_top_module}_v.so'

  #---------------------
  # Command generation
  #---------------------

  def create_vl_cmd( s ):
    top_module  = f"--top-module {s.translated_top_module}"
    src         = s.translated_source_file
    mk_dir      = f"--Mdir {s.vl_mk_dir}"
    # flist       = "" if s.is_default("v_flist") else \
    #               f"-f {s.v_flist}"
    # We don't need this since all v_libs appear in the translation result
    # vlibs       = "" if not s.v_libs else \
    #               " ".join([f"-v {lib}" for lib in s.v_libs])
    vlibs       = ""
    include     = "" if not s.v_include else \
                  " ".join("-I" + path for path in s.v_include)
    en_assert   = "--assert" if s.vl_enable_assert else ""

    # Always verilator -O3 and unroll because -O0 may lead to this error:
    # -Info: Command Line disabled gate optimization with -Og/-O0.  This may cause ordering problems
    opt_level   = "-O3"
    loop_unroll = "--unroll-count 1000000"
    stmt_unroll = "--unroll-stmts 1000000"
    trace       = "--trace" if s.vl_trace else ""
    coverage    = "--coverage" if s.vl_coverage else ""
    line_cov    = "--coverage-line" if s.vl_line_coverage else ""
    toggle_cov  = "--coverage-toggle" if s.vl_toggle_coverage else ""
    warnings    = s._create_vl_warning_cmd()

    all_opts = [
      top_module, mk_dir, include, en_assert, opt_level, loop_unroll,
      # stmt_unroll, trace, warnings, flist, src, coverage,
      stmt_unroll, trace, warnings, src, vlibs, coverage,
      line_cov, toggle_cov,
    ]

    return f"verilator --cc {' '.join(opt for opt in all_opts if opt)}"

  def create_cc_cmd( s ):
    # Shunning: GCC has a family of optimizations on the SSA tree -ftree-***.
    # It belongs to -O1, and takes a lot of time to execute when I compile the verilated C++.
    # However, the compiled code has worse performance after applying those tree optimizations.
    # I guess it's because verilator already emits very regular and low-level code in the form of C++.
    # Also gcc -O1, -O2, -O3 result in the same compilation time after all optimization flags are turned off
    # gcc -O0 still has shorter compilation time than O1-3 with all flags off
    # My guess is that -O1 has some hidden optimization that the user cannot turn off, and -O2-3 inherit that part
    # the difference between O1-3 is just flags
    # so right now I'm using "-O1 with all options off" to get the fastest compilation time and good performance
    # !!! -fno-tree-forwprop is removed due to:
    # In file included from /usr/include/fcntl.h:290:0,
    #                   from /usr/local/share/verilator/include/verilated_vcd_c.cpp:29:
    # In function ‘int open(const char*, int, ...)’,
    #     inlined from ‘virtual bool VerilatedVcdFile::open(const string&)’ at
    # /usr/local/share/verilator/include/verilated_vcd_c.cpp:116:18:
    # /usr/include/x86_64-linux-gnu/bits/fcntl2.h:44:26: error: call to
    # ‘__open_too_many_args’ declared with attribute error: open can be called either
    # with 2 or 3 arguments, not more
    #       __open_too_many_args ();
    #       ~~~~~~~~~~~~~~~~~~~~~^~
    # /usr/include/x86_64-linux-gnu/bits/fcntl2.h:50:24: error: call to
    # ‘__open_missing_mode’ declared with attribute error: open with O_CREAT or
    # O_TMPFILE in second argument needs 3 arguments
    #     __open_missing_mode ();
    #     ~~~~~~~~~~~~~~~~~~~~^~
    # Shunning (7/7/2020): Use O0 here because O1 will lead to corruptions
    # in the CFFI-managed shared library pools. Basically what happened
    # is if we run two tests with the same component name but different
    # contents in a row, the second one uses the first library instead
    # of the second ...
    # (7/9/2020): Use -O0 by default so that normally the tests are super fast and don't corrupt cffi,
    # but when the user gives a "fast" flag, it uses -O1.
    if s.fast:
      c_flags = "-O1 -fno-guess-branch-probability -fno-reorder-blocks -fno-if-conversion -fno-if-conversion2 -fno-dce -fno-delayed-branch -fno-dse -fno-auto-inc-dec -fno-branch-count-reg -fno-combine-stack-adjustments -fno-cprop-registers -fno-forward-propagate -fno-inline-functions-called-once -fno-ipa-profile -fno-ipa-pure-const -fno-ipa-reference -fno-move-loop-invariants -fno-omit-frame-pointer -fno-split-wide-types -fno-tree-bit-ccp -fno-tree-ccp -fno-tree-ch -fno-tree-coalesce-vars -fno-tree-copy-prop -fno-tree-dce -fno-tree-dominator-opts -fno-tree-dse -fno-tree-fre -fno-tree-phiprop -fno-tree-pta -fno-tree-scev-cprop -fno-tree-sink -fno-tree-slsr -fno-tree-sra -fno-tree-ter -fno-tree-reassoc -fPIC -fno-gnu-unique -shared"
    else:
      c_flags = "-O0 -fno-guess-branch-probability -fno-reorder-blocks -fno-if-conversion -fno-if-conversion2 -fno-dce -fno-delayed-branch -fno-dse -fno-auto-inc-dec -fno-branch-count-reg -fno-combine-stack-adjustments -fno-cprop-registers -fno-forward-propagate -fno-inline-functions-called-once -fno-ipa-profile -fno-ipa-pure-const -fno-ipa-reference -fno-move-loop-invariants -fno-omit-frame-pointer -fno-split-wide-types -fno-tree-bit-ccp -fno-tree-ccp -fno-tree-ch -fno-tree-coalesce-vars -fno-tree-copy-prop -fno-tree-dce -fno-tree-dominator-opts -fno-tree-dse -fno-tree-fre -fno-tree-phiprop -fno-tree-pta -fno-tree-scev-cprop -fno-tree-sink -fno-tree-slsr -fno-tree-sra -fno-tree-ter -fno-tree-reassoc -fPIC -fno-gnu-unique -shared"

    if not s.is_default("c_flags"):
      c_flags += f" {expand(s.c_flags)}"

    c_include_path = " ".join("-I"+p for p in s._get_all_includes() if p)
    out_file = s.get_shared_lib_path()
    c_src_files = " ".join(s._get_c_src_files())
    ld_flags = expand(s.ld_flags)
    ld_libs = s.ld_libs
    coverage = "-DVM_COVERAGE" if s.vl_coverage or \
                                  s.vl_line_coverage or \
                                  s.vl_toggle_coverage else ""
    return f"g++ {c_flags} {c_include_path} {ld_flags}"\
           f" -o {out_file} {c_src_files} {ld_libs} {coverage}"

  def vprint( s, msg, nspaces = 0, use_fill = False ):
    if s.verbose:
      if use_fill:
        print(indent(fill(msg), " "*nspaces))
      else:
        print(indent(msg, " "*nspaces))

  #---------------------
  # Internal helpers
  #---------------------

  def _create_vl_warning_cmd( s ):
    lint = "" if s.is_default("vl_W_lint") else "--Wno-lint"
    style = "" if s.is_default("vl_W_style") else "--Wno-style"
    fatal = "" if s.is_default("vl_W_fatal") else "--Wno-fatal"
    wno = " ".join(f"--Wno-{w}" for w in s.vl_Wno_list)
    return " ".join(w for w in [lint, style, fatal, wno] if w)

  def _get_all_includes( s ):
    includes = copy.copy(s.c_include_path)

    # Try to obtain verilator include path either from environment variable
    # or from `pkg-config`
    vl_include_dir = os.environ.get("PYMTL_VERILATOR_INCLUDE_DIR")
    if vl_include_dir is None:
      get_dir_cmd = ["pkg-config", "--variable=includedir", "verilator"]
      try:
        vl_include_dir = \
            subprocess.check_output(get_dir_cmd, stderr = subprocess.STDOUT).strip()
        vl_include_dir = vl_include_dir.decode('ascii')
      except OSError as e:
        vl_include_dir_msg = \
"""\
Cannot locate the include directory of verilator. Please make sure either \
$PYMTL_VERILATOR_INCLUDE_DIR is set or `pkg-config` has been configured properly!
"""
        raise OSError(fill(vl_include_dir_msg)) from e

    # Add verilator include path
    s.vl_include_dir = vl_include_dir
    includes += [vl_include_dir, vl_include_dir + "/vltstd"]

    return includes

  def _get_c_src_files( s ):
    top_module = s.translated_top_module
    vl_mk_dir = s.vl_mk_dir
    vl_class_mk = f"{vl_mk_dir}/V{top_module}_classes.mk"

    # Add C wrapper
    o0 = []
    o1 = copy.copy(s.c_srcs) + [ s.get_c_wrapper_path() ]

    # Add files listed in class makefile
    with open(vl_class_mk) as class_mk:
      all_lines = class_mk.readlines()
      o1 += s._get_srcs_from_vl_class_mk( all_lines, vl_mk_dir, "VM_CLASSES_FAST")
      o1 += s._get_srcs_from_vl_class_mk( all_lines, vl_mk_dir, "VM_SUPPORT_FAST")
      o1 += s._get_srcs_from_vl_class_mk( all_lines, s.vl_include_dir, "VM_GLOBAL_FAST")

      o0 += s._get_srcs_from_vl_class_mk( all_lines, vl_mk_dir, "VM_CLASSES_SLOW")
      o0 += s._get_srcs_from_vl_class_mk( all_lines, vl_mk_dir, "VM_SUPPORT_SLOW")
      o0 += s._get_srcs_from_vl_class_mk( all_lines, s.vl_include_dir, "VM_GLOBAL_SLOW")

    with open(f"{top_module}_v__ALL_pickled.cpp", 'w') as out:

      if not s.fast:
        out.write('\n'.join( [ f'#include "{x}"' for x in o1 + o0 ]))

      else:
        if o1:
          out.write('\n'.join( [ f'#include "{x}"' for x in o1 ]))

        out.write('\n')

        if o0:
        # Apply no optimization for SLOW files
          out.write('\n#pragma GCC push_options')
          out.write('\n#pragma GCC optimize ("O0")\n')
          out.write('\n'.join( [ f'#include "{x}"' for x in o0 ]))
          out.write('\n#pragma GCC pop_options\n')

    return [ f"{top_module}_v__ALL_pickled.cpp" ]

  def _get_srcs_from_vl_class_mk( s, all_lines, path, label ):
    """Return all files under `path` directory in `label` section of `mk`."""
    srcs, found = [], False
    for line in all_lines:
      if line.startswith(label):
        found = True
      elif found:
        if line.strip() == "":
          found = False
        else:
          file_name = line.strip()[:-2]
          srcs.append( path + "/" + file_name + ".cpp" )
    return srcs



# from .EnqDeqIfc import DeqIfcFL, DeqIfcRTL, EnqIfcFL, EnqIfcRTL
from .get_give_ifcs import GetIfcFL, GetIfcRTL, GiveIfcFL, GiveIfcRTL
from .master_minion_ifcs import MasterIfcCL, MasterIfcRTL, MinionIfcCL, MinionIfcRTL
from .send_recv_ifcs import (
    RecvCL2SendRTL,
    RecvIfcFL,
    RecvIfcRTL,
    RecvRTL2SendCL,
    SendIfcFL,
    SendIfcRTL,
)
from .xcel_ifcs import (
    XcelMasterIfcCL,
    XcelMasterIfcFL,
    XcelMasterIfcRTL,
    XcelMinionIfcCL,
    XcelMinionIfcFL,
    XcelMinionIfcRTL,
)
from .XcelMsg import XcelMsgType, mk_xcel_msg, mk_xcel_req_msg, mk_xcel_resp_msg



#=========================================================================
# BehavioralRTLIRTypeCheckL3Pass.py
#=========================================================================
# Author : Peitian Pan
# Date   : March 30, 2019
"""Provide L3 behavioral RTLIR type check pass."""

from pymtl3.passes.rtlir.errors import PyMTLTypeError
from pymtl3.passes.rtlir.rtype import RTLIRDataType as rdt
from pymtl3.passes.rtlir.rtype import RTLIRType as rt

from .BehavioralRTLIRTypeCheckL2Pass import (
    BehavioralRTLIRTypeCheckL2Pass,
    BehavioralRTLIRTypeCheckVisitorL2,
    BehavioralRTLIRTypeEnforcerL2,
)


class BehavioralRTLIRTypeCheckL3Pass( BehavioralRTLIRTypeCheckL2Pass ):
  def get_visitor_class( s ):
    return BehavioralRTLIRTypeCheckVisitorL3

class BehavioralRTLIRTypeCheckVisitorL3( BehavioralRTLIRTypeCheckVisitorL2 ):
  def __init__( s, component, freevars, accessed, tmpvars, rtlir_getter ):
    super().__init__( component, freevars, accessed, tmpvars, rtlir_getter )
    s.type_expect[ 'Attribute' ] = (
      ( 'value', (rt.Component, rt.Signal),
                 'the base of an attribute must be one of: component, signal!' ),
    )

  def get_enforce_visitor( s ):
    return BehavioralRTLIRTypeEnforcerL3

  def _visit_Assign_single_target( s, node, target, i ):
    try:
      rhs_type = node.value.Type.get_dtype()
      lhs_type = target.Type.get_dtype()
    except AttributeError:
      rhs_type = None
      lhs_type = None

    l_is_struct = isinstance( lhs_type, rdt.Struct )
    r_is_struct = isinstance( rhs_type, rdt.Struct )

    # At L3 we check if either LHS or RHS is of BitStruct type.
    if l_is_struct or r_is_struct:
      if l_is_struct and r_is_struct:
        # Both sides are of struct type. Type check only if both sides
        # are of the _same_ struct type.
        if lhs_type.get_name() != rhs_type.get_name():
          raise PyMTLTypeError( s.blk, node.ast,
            f'LHS and RHS of assignment should have the same type (LHS target#{i+1} of {lhs_type} vs {rhs_type})!' )
      else:
        # There should be one side of struct type and the other of vector type.
        struct_type, vector_type = lhs_type, rhs_type
        if r_is_struct:
          struct_type, vector_type = rhs_type, lhs_type
        if not isinstance( vector_type, rdt.Vector ):
          raise PyMTLTypeError( s.blk, node.ast,
            f'LHS and RHS of assignment should have agreeable types (LHS target#{i+1} of {lhs_type} vs {rhs_type})!' )

        # Type check only if both sides have the same bitwidth.
        is_rhs_reinterpretable = not node.value._is_explicit
        struct_nbits, vector_nbits = struct_type.get_length(), vector_type.get_length()

        # If RHS is an int literal try to enforce the correct bitwidth.
        if not r_is_struct and is_rhs_reinterpretable and struct_nbits != vector_nbits:
          s.enforcer.enter( s.blk, rt.NetWire(rdt.Vector(struct_nbits)), node.value )

        if l_is_struct:
          vector_nbits = node.value.Type.get_dtype().get_length()

        if struct_nbits != vector_nbits:
          if l_is_struct:
            lnbits, rnbits = struct_nbits, vector_nbits
          else:
            lnbits, rnbits = vector_nbits, struct_nbits

          raise PyMTLTypeError( s.blk, node.ast,
            f'LHS and RHS of assignment should have the same bitwidth (LHS target#{i+1} of {lhs_type} ({lnbits} bits) vs {rhs_type} ({rnbits} bits))!' )

    else:
      super()._visit_Assign_single_target( node, target, i )

  def visit_Attribute( s, node ):
    if isinstance( node.value.Type, rt.Signal ):
      dtype = node.value.Type.get_dtype()
      if not isinstance( dtype, rdt.Struct ):
        raise PyMTLTypeError( s.blk, node.ast,
          'attribute base should be a struct signal!'
        )
      if not dtype.has_property( node.attr ):
        raise PyMTLTypeError( s.blk, node.ast,
          f'{dtype.get_name()} does not have field {node.attr}!' )
      dtype = dtype.get_property( node.attr )
      if isinstance( node.value.Type, rt.Port ):
        rtype = rt.Port( node.value.Type.get_direction(), dtype )
      elif isinstance( node.value.Type, rt.Wire ):
        rtype = rt.Wire( dtype )
      elif isinstance( node.value.Type, rt.Const ):
        obj = node.value.Type.get_object()
        if obj is None:
          rtype = rt.Const( dtype )
        else:
          try:
            rtype = rt.Const( dtype, getattr( obj, node.attr ) )
          except AttributeError:
            rtype = rt.Const( dtype )
      else:
        raise PyMTLTypeError( s.blk, node.ast,
          f'unrecognized signal type {node.value.Type}!' )
      node.Type = rtype
      dtype = node.Type.get_dtype()
      # Only allow to be re-interpreted if this is a constant vector attribute
      if isinstance( node.Type, rt.Const ) and isinstance( dtype, rdt.Vector ):
        node._is_explicit = dtype.is_explicit()
      else:
        node._is_explicit = True

    else:
      super().visit_Attribute( node )

  def visit_StructInst( s, node ):
    cls = node.struct

    dtype = rdt.get_rtlir_dtype( cls() )
    all_properties = dtype.get_all_properties()

    if len( all_properties ) != len( node.values ):
      raise PyMTLTypeError( s.blk, node.ast,
        f"BitStruct {cls.__name__} has {len(all_properties)} fields but only {len(node.values)} arguments are given!" )

    all_types = zip( node.values, list(all_properties.items()) )

    for idx, ( value, ( name, field ) ) in enumerate( all_types ):
      s.visit( value )
      # Expect each argument to be a signal
      if not isinstance( value.Type, rt.Signal ):
        raise PyMTLTypeError( s.blk, node.ast,
          f"argument #{idx+1} has type {value.Type} but not a signal!" )
      v_dtype = value.Type.get_dtype()
      is_field_reinterpretable = not value._is_explicit
      # Expect each argument to have data type which corresponds to the field
      if v_dtype != field:
        if is_field_reinterpretable:
          target_nbits = field.get_length()
          s.enforcer.enter( s.blk, rt.NetWire(rdt.Vector(target_nbits)), value )
        else:
          raise PyMTLTypeError( s.blk, node.ast,
            f"Expected argument#{idx+1} ( field {name} ) to have type {field}, but got {v_dtype}." )

    node.Type = rt.Const( dtype )
    node._is_explicit = True

#-------------------------------------------------------------------------
# Enforce types for all terms whose types are inferred (implicit)
#-------------------------------------------------------------------------

class BehavioralRTLIRTypeEnforcerL3( BehavioralRTLIRTypeEnforcerL2 ):
  pass



template = \
'''
#=========================================================================
# V{component_name}_v.py
#=========================================================================
"""Provide a template of PyMTL wrapper to import verilated models.

This wrapper makes a Verilator-generated C++ model appear as if it were a
normal PyMTL model. This template is based on PyMTL v2.
"""

import os

from cffi import FFI

from pymtl3.datatypes import *
from pymtl3.dsl import Component, connect, InPort, OutPort, Wire, update, update_ff

#-------------------------------------------------------------------------
# {component_name}
#-------------------------------------------------------------------------

class {component_name}( Component ):
  id_ = 0

  def __init__( s, *args, **kwargs ):
    s._finalization_count = 0

    # initialize FFI, define the exposed interface
    s.ffi = FFI()
    s.ffi.cdef("""
      typedef struct {{

        // Exposed port interface
{port_cdefs}

        // Verilator model
        void * model;

      }} V{component_name}_t;

      V{component_name}_t * create_model( const char * );
      void destroy_model( V{component_name}_t *);
      void comb_eval( V{component_name}_t * );
      void seq_eval( V{component_name}_t * );
      void assert_en( bool en );
      {trace_c_def}

    """)

    # Print the modification time stamp of the shared lib
    # print 'Modification time of {{}}: {{}}'.format(
    #   '{lib_file}', os.path.getmtime( './{lib_file}' ) )

    # Import the shared library containing the model. We defer
    # construction to the elaborate_logic function to allow the user to
    # set the vcd_file.
    s._ffi_inst = s.ffi.dlopen('./{lib_file}')

    # increment instance count
    {component_name}.id_ += 1

  def finalize( s ):
    """Finalize the imported component.

    This method closes the shared library opened through CFFI. If an imported
    component is not finalized explicitly (i.e. if you rely on GC to collect a
    no longer used imported component), importing a component with the same
    name before all previous imported components are GCed might lead to
    confusing behaviors. This is because once opened, the shared lib
    is cached by the OS until the OS reference counter for this lib reaches
    0 (you can decrement the reference counter by calling `dl_close()` syscall).

    Fortunately real designs tend to always have the same shared lib corresponding
    to the components with the same name. If you are doing translation testing and
    use the same component class name even if they refer to different designs,
    you might need to call `imported_object.finalize()` at the end of each test
    to ensure correct behaviors.
    """
    assert s._finalization_count == 0,\
      'Imported component can only be finalized once!'
    s._finalization_count += 1
    s._ffi_inst.destroy_model( s._ffi_m )
    s.ffi.dlclose( s._ffi_inst )
    s.ffi = None
    s._ffi_inst = None

  def __del__( s ):
    if s._finalization_count == 0:
      s._finalization_count += 1
      s._ffi_inst.destroy_model( s._ffi_m )
      s.ffi.dlclose( s._ffi_inst )
      s.ffi = None
      s._ffi_inst = None

  def construct( s, *args, **kwargs ):
    # Set up the VCD file name
    verilator_vcd_file = ""
    if {dump_vcd}:
      if {has_vl_trace_filename}:
        verilator_vcd_file = "{vl_trace_filename}.verilator1.vcd"
      else:
        verilator_vcd_file = "{component_name}.verilator1.vcd"

    # Convert string to `bytes` which is required by CFFI on python 3
    verilator_vcd_file = verilator_vcd_file.encode('ascii')

    # Construct the model
    s._ffi_m = s._ffi_inst.create_model( s.ffi.new("char[]", verilator_vcd_file) )

    # Buffer for line tracing
    s._line_trace_str = s.ffi.new('char[512]')
    s._convert_string = s.ffi.string

    # Use non-attribute varialbe to reduce CPython bytecode count
    _ffi_m = s._ffi_m
    _ffi_inst_comb_eval = s._ffi_inst.comb_eval
    _ffi_inst_seq_eval  = s._ffi_inst.seq_eval

    # declare the port interface
{port_defs}

    # update blocks that converts ffi interface to/from pymtl ports
{structs_input}
{structs_output}

    @update
    def comb_upblk():

      # Set inputs
{set_comb_input}

      _ffi_inst_comb_eval( _ffi_m )

      # Write all outputs
{set_comb_output}

    @update_ff
    def seq_upblk():
      # seq_eval will automatically tick clock in C land
      _ffi_inst_seq_eval( _ffi_m )

  def assert_en( s, en ):
    # TODO: for verilator, any assertion failure will cause the C simulator
    # to abort, which results in a Python internal error. A better approach
    # is to throw a Python exception at the time of assertion failure.
    # Verilator allows user-defined `stop` function which is called when
    # the simulation is expected to stop due to various reasons. We might
    # be able to raise a Python exception through Python C API (although
    # at this moment I'm not sure if the C API's are compatible between
    # PyPy and CPython).
    assert isinstance( en, bool )
    s._ffi_inst.assert_en( en )

  def line_trace( s ):
    if {external_trace}:
      s._ffi_inst.trace( s._ffi_m, s._line_trace_str )
      return s._convert_string( s._line_trace_str ).decode('ascii')
    else:
{line_trace}

  def internal_line_trace( s ):
{in_line_trace}
'''



"""
========================================================================
SinkRTL
========================================================================
Test sinks with RTL interfaces.

Author : Shunning Jiang
  Date : Feb 12, 2021
"""

from pymtl3 import *
from .ifcs import RecvIfcRTL


class PyMTLTestSinkError( Exception ): pass

#-------------------------------------------------------------------------
# TestSinkRTL
#-------------------------------------------------------------------------

class SinkRTL( Component ):

  def construct( s, Type, msgs, initial_delay=0, interval_delay=0,
                 arrival_time=None, cmp_fn=lambda a, b : a == b ):

    # Interface

    s.recv = RecvIfcRTL( Type )

    # Data

    # [msgs] and [arrival_time] must have the same length.
    if arrival_time is None:
      s.arrival_time = None
    else:
      assert len( msgs ) == len( arrival_time )
      s.arrival_time = list( arrival_time )

    s.idx          = 0
    s.count        = 0
    s.cycle_count  = 0
    s.msgs         = list( msgs )

    s.error_msg    = ''

    s.received = False
    s.all_msg_recved = False
    s.done_flag      = False

    @update_ff
    def up_sink():
      # Raise exception at the start of next cycle so that the errored
      # line trace gets printed out
      if s.error_msg:
        raise PyMTLTestSinkError( s.error_msg )

      # Tick one more cycle after all message is received so that the
      # exception gets thrown
      if s.all_msg_recved:
        s.done_flag = True

      if s.idx >= len(s.msgs):
        s.all_msg_recved = True

      s.received = False

      if s.reset:
        s.cycle_count = 0

        s.idx = 0
        s.count = initial_delay
        s.recv.rdy <<= (s.idx < len(s.msgs)) & (s.count == 0)

      else:
        s.cycle_count += 1

        # This means at least previous cycle count = 0
        if s.recv.val & s.recv.rdy:
          msg = s.recv.msg

          # Sanity check
          if s.idx >= len(s.msgs):
            s.error_msg = ( 'Test Sink received more msgs than expected!\n'
                           f'Received : {msg}' )

          else:
            # Check correctness first
            if not cmp_fn( msg, s.msgs[ s.idx ] ):
              s.error_msg = (
                f'Test sink {s} received WRONG message!\n'
                f'Expected : { s.msgs[ s.idx ] }\n'
                f'Received : { msg }'
              )

            # Check timing if performance regeression is turned on
            elif s.arrival_time and s.cycle_count > s.arrival_time[ s.idx ]:
              s.error_msg = (
                f'Test sink {s} received message LATER than expected!\n'
                f'Expected msg : {s.msgs[ s.idx ]}\n'
                f'Expected at  : {s.arrival_time[ s.idx ]}\n'
                f'Received msg : {msg}\n'
                f'Received at  : {s.cycle_count}'
              )

          s.idx += 1
          s.count = interval_delay

        if s.count > 0:
          s.count -= 1
          s.recv.rdy <<= 0
        else: # s.count == 0
          s.recv.rdy <<= (s.idx < len(s.msgs))

  def done( s ):
    return s.done_flag

  # Line trace

  def line_trace( s ):
    return f"{s.recv}"



#=========================================================================
# VNameMangle_test.py
#=========================================================================
# Author : Peitian Pan
# Date   : May 30, 2019
"""Test the SystemVerilog name mangling."""

from pymtl3.datatypes import Bits1, Bits32, bitstruct
from pymtl3.dsl import Component, InPort, Interface, OutPort
from pymtl3.passes.backends.verilog.util.utility import gen_mapped_ports
from pymtl3.passes.rtlir import RTLIRDataType as rdt
from pymtl3.passes.rtlir import RTLIRType as rt
from pymtl3.passes.rtlir.util.test_utility import do_test


def local_do_test( m ):
  m.elaborate()
  result = gen_mapped_ports( m, {} )
  assert result == m._ref_ports

def test_port_single( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( Bits32 )
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = a._ref_ports
  do_test( a )

def test_port_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( Bits32 ) for _ in range( 3 ) ]
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Array([3], rt.Port('input', rdt.Vector(32))), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_[0]'], 'in___0', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[1]'], 'in___1', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[2]'], 'in___2', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  do_test( a )

def test_port_2d_array( do_test ):
  class A( Component ):
    def construct( s ):
      s.in_ = [ [ InPort( Bits32 ) for _ in range(2) ] for _ in range(3) ]
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Array( [3, 2], rt.Port('input', rdt.Vector(32)) ), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_[0][0]'], 'in___0__0', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[0][1]'], 'in___0__1', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[1][0]'], 'in___1__0', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[1][1]'], 'in___1__1', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[2][0]'], 'in___2__0', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['in_[2][1]'], 'in___2__1', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  do_test( a )

def test_struct_port_single( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = InPort( struct )
  a = A()
  st = rdt.Struct(struct, {'bar':rdt.Vector(32), 'foo':rdt.Vector(32)})
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Port('input', st ), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_.bar'], 'in___bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_.foo'], 'in___foo', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ]
  do_test( a )

def test_struct_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: Bits32
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  st = rdt.Struct(struct, {'bar':rdt.Vector(32), 'foo':rdt.Vector(32)})
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Array([2], rt.Port('input', st)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_[0].bar'], 'in___0__bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo'], 'in___0__foo', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].bar'], 'in___1__bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo'], 'in___1__foo', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  do_test( a )

def test_packed_array_port_array( do_test ):
  @bitstruct
  class struct:
    bar: Bits32
    foo: [ [ Bits32 ] * 2 ] * 3
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  foo = rdt.PackedArray([3,2], rdt.Vector(32))
  st = rdt.Struct(struct, {'bar':rdt.Vector(32), 'foo':foo})
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Array([2], rt.Port('input', st)), 0),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_[0].bar'], 'in___0__bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo[0][0]'], 'in___0__foo__0__0', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo[0][1]'], 'in___0__foo__0__1', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo[1][0]'], 'in___0__foo__1__0', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo[1][1]'], 'in___0__foo__1__1', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo[2][0]'], 'in___0__foo__2__0', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].foo[2][1]'], 'in___0__foo__2__1', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].bar'], 'in___1__bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo[0][0]'], 'in___1__foo__0__0', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo[0][1]'], 'in___1__foo__0__1', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo[1][0]'], 'in___1__foo__1__0', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo[1][1]'], 'in___1__foo__1__1', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo[2][0]'], 'in___1__foo__2__0', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].foo[2][1]'], 'in___1__foo__2__1', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  do_test( a )

def test_nested_struct( do_test ):
  @bitstruct
  class inner_struct:
    foo: Bits32
  @bitstruct
  class struct:
    bar: Bits32
    inner: inner_struct
  class A( Component ):
    def construct( s ):
      s.in_ = [ InPort( struct ) for _ in range(2) ]
  a = A()
  inner = rdt.Struct(inner_struct, {'foo':rdt.Vector(32)})
  st = rdt.Struct(struct, {'bar':rdt.Vector(32), 'inner':inner})
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_'], 'in_', rt.Array([2], rt.Port('input', st )), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['in_[0].bar'], 'in___0__bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[0].inner.foo'], 'in___0__inner__foo', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].bar'], 'in___1__bar', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['in_[1].inner.foo'], 'in___1__inner__foo', rt.Port('input', rdt.Vector(32) ), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  do_test( a )

def test_interface( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = Ifc()
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc.msg'], 'ifc__msg', rt.Port('input', rdt.Vector(32)), 0 ),
    ( ['ifc.rdy'], 'ifc__rdy', rt.Port('output', rdt.Vector(1)), 0 ),
    ( ['ifc.val'], 'ifc__val', rt.Port('input', rdt.Vector(1)), 0 ),
  ]
  a._ref_ports_yosys = a._ref_ports
  do_test( a )

def test_interface_array( do_test ):
  class Ifc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc[0].msg', 'ifc[1].msg'], 'ifc__msg', rt.Array([2], rt.Port('input', rdt.Vector(32))), 1 ),
    ( ['ifc[0].rdy', 'ifc[1].rdy'], 'ifc__rdy', rt.Array([2], rt.Port('output', rdt.Vector(1))), 1 ),
    ( ['ifc[0].val', 'ifc[1].val'], 'ifc__val', rt.Array([2], rt.Port('input', rdt.Vector(1))), 1 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc[0].msg'], 'ifc__0__msg', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[0].rdy'], 'ifc__0__rdy', rt.Port('output', rdt.Vector(1)), 1 ),
    ( ['ifc[0].val'], 'ifc__0__val', rt.Port('input', rdt.Vector(1)), 1 ),
    ( ['ifc[1].msg'], 'ifc__1__msg', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[1].rdy'], 'ifc__1__rdy', rt.Port('output', rdt.Vector(1)), 1 ),
    ( ['ifc[1].val'], 'ifc__1__val', rt.Port('input', rdt.Vector(1)), 1 ),
  ]
  do_test( a )

def test_nested_interface( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = InPort( Bits32 )
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc[0].ctrl_bar', 'ifc[1].ctrl_bar'], 'ifc__ctrl_bar', rt.Array([2], rt.Port('input', rdt.Vector(32))), 1 ),
    ( ['ifc[0].ctrl_foo', 'ifc[1].ctrl_foo'], 'ifc__ctrl_foo', rt.Array([2], rt.Port('output', rdt.Vector(32))), 1 ),
    ( ['ifc[0].valrdy_ifc.msg', 'ifc[1].valrdy_ifc.msg'], 'ifc__valrdy_ifc__msg', rt.Array([2], rt.Port('input', rdt.Vector(32))), 1 ),
    ( ['ifc[0].valrdy_ifc.rdy', 'ifc[1].valrdy_ifc.rdy'], 'ifc__valrdy_ifc__rdy', rt.Array([2], rt.Port('output', rdt.Vector(1))), 1 ),
    ( ['ifc[0].valrdy_ifc.val', 'ifc[1].valrdy_ifc.val'], 'ifc__valrdy_ifc__val', rt.Array([2], rt.Port('input', rdt.Vector(1))), 1 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc[0].ctrl_bar'], 'ifc__0__ctrl_bar', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[0].ctrl_foo'], 'ifc__0__ctrl_foo', rt.Port('output', rdt.Vector(32)), 1 ),
    ( ['ifc[0].valrdy_ifc.msg'], 'ifc__0__valrdy_ifc__msg', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[0].valrdy_ifc.rdy'], 'ifc__0__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)), 1 ),
    ( ['ifc[0].valrdy_ifc.val'], 'ifc__0__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)), 1 ),
    ( ['ifc[1].ctrl_bar'], 'ifc__1__ctrl_bar', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[1].ctrl_foo'], 'ifc__1__ctrl_foo', rt.Port('output', rdt.Vector(32)), 1 ),
    ( ['ifc[1].valrdy_ifc.msg'], 'ifc__1__valrdy_ifc__msg', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[1].valrdy_ifc.rdy'], 'ifc__1__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)), 1 ),
    ( ['ifc[1].valrdy_ifc.val'], 'ifc__1__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)), 1 ),
  ]
  do_test( a )

def test_nested_interface_port_array( do_test ):
  class InnerIfc( Interface ):
    def construct( s ):
      s.msg = [ InPort( Bits32 ) for _ in range(2) ]
      s.val = InPort( Bits1 )
      s.rdy = OutPort( Bits1 )
  class Ifc( Interface ):
    def construct( s ):
      s.valrdy_ifc = InnerIfc()
      s.ctrl_bar = InPort( Bits32 )
      s.ctrl_foo = OutPort( Bits32 )
  class A( Component ):
    def construct( s ):
      s.ifc = [ Ifc() for _ in range(2) ]
  a = A()
  a._ref_ports = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc[0].ctrl_bar', 'ifc[1].ctrl_bar'], 'ifc__ctrl_bar', rt.Array([2], rt.Port('input', rdt.Vector(32))), 1 ),
    ( ['ifc[0].ctrl_foo', 'ifc[1].ctrl_foo'], 'ifc__ctrl_foo', rt.Array([2], rt.Port('output', rdt.Vector(32))), 1 ),
    ( ['ifc[0].valrdy_ifc.msg', 'ifc[1].valrdy_ifc.msg'], 'ifc__valrdy_ifc__msg', rt.Array([2, 2], rt.Port('input', rdt.Vector(32))), 1 ),
    ( ['ifc[0].valrdy_ifc.rdy', 'ifc[1].valrdy_ifc.rdy'], 'ifc__valrdy_ifc__rdy', rt.Array([2], rt.Port('output', rdt.Vector(1))), 1 ),
    ( ['ifc[0].valrdy_ifc.val', 'ifc[1].valrdy_ifc.val'], 'ifc__valrdy_ifc__val', rt.Array([2], rt.Port('input', rdt.Vector(1))), 1 ),
  ]
  a._ref_ports_yosys = [
    ( ['clk'], 'clk', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['reset'], 'reset', rt.Port('input', rdt.Vector(1)), 0 ),
    ( ['ifc[0].ctrl_bar'], 'ifc__0__ctrl_bar', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[0].ctrl_foo'], 'ifc__0__ctrl_foo', rt.Port('output', rdt.Vector(32)), 1 ),
    ( ['ifc[0].valrdy_ifc.msg[0]'], 'ifc__0__valrdy_ifc__msg__0', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[0].valrdy_ifc.msg[1]'], 'ifc__0__valrdy_ifc__msg__1', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[0].valrdy_ifc.rdy'], 'ifc__0__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)), 1 ),
    ( ['ifc[0].valrdy_ifc.val'], 'ifc__0__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)), 1 ),
    ( ['ifc[1].ctrl_bar'], 'ifc__1__ctrl_bar', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[1].ctrl_foo'], 'ifc__1__ctrl_foo', rt.Port('output', rdt.Vector(32)), 1 ),
    ( ['ifc[1].valrdy_ifc.msg[0]'], 'ifc__1__valrdy_ifc__msg__0', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[1].valrdy_ifc.msg[1]'], 'ifc__1__valrdy_ifc__msg__1', rt.Port('input', rdt.Vector(32)), 1 ),
    ( ['ifc[1].valrdy_ifc.rdy'], 'ifc__1__valrdy_ifc__rdy', rt.Port('output', rdt.Vector(1)), 1 ),
    ( ['ifc[1].valrdy_ifc.val'], 'ifc__1__valrdy_ifc__val', rt.Port('input', rdt.Vector(1)), 1 ),
  ]
  do_test( a )



#=========================================================================
# DynamicSchedulePass_test.py
#=========================================================================
#
# Author : Shunning Jiang
# Date   : Apr 19, 2019

from pymtl3.dsl import Component, Wire, update_ff

from ... import DefaultPassGroup
from ...backends.verilog import *
from ..AddDebugSignalPass import AddDebugSignalPass


def test_bring_up_non_top_debug_signal():

  class Mult(Component):
    def construct( s, init=0 ):
      s.en = Wire(32)
      @update_ff
      def ff():
        if s.reset:
          s.en <<= init
        else:
          s.en <<= s.en ^ 1

  class Dpath(Component):
    def construct( s ):
      s.mult = Mult()

  class Core(Component):
    def construct( s ):
      s.dpath = Dpath()

  class Tile(Component):
    def construct( s ):
      s.core = Core()

  class CoreChip(Component):
    def construct( s ):
      s.tile0 = Tile()
      # s.tile1 = Tile()

  class FullChip(Component):
    def construct( s ):
      s.corechip = CoreChip()

  class Top(Component):
    def construct( s ):
      s.fullchip = FullChip()

    def line_trace( s ):
      return str(s.debug_0)

  top1 = Top()
  top1.set_param( "top.fullchip.corechip.tile0.core.dpath.mult.construct", init=0x100 )
  # top.set_param( "top.fullchip.corechip.tile1.core.dpath.mult.construct", init=0x888 )
  top1.elaborate()

  AddDebugSignalPass()( top1, [ "top.fullchip.corechip.tile0.core.dpath.mult.en",
                              # "top.fullchip.corechip.tile1.core.dpath.mult.en",
                              ])

  top1.apply( DefaultPassGroup() )
  top1.sim_reset()

  for i in range(10):
    top1.sim_tick()

  top2 = Top()
  top2.set_param( "top.fullchip.corechip.tile0.core.dpath.mult.construct", init=0x100 )
  # top2.set_param( "top.fullchip.corechip.tile1.core.dpath.mult.construct", init=0x888 )
  top2.elaborate()

  AddDebugSignalPass()( top2, [ "top.fullchip.corechip.tile0.core.dpath.mult.en",
                              # "top.fullchip.corechip.tile1.core.dpath.mult.en",
                              ])

  top2.set_metadata( VerilogTranslationImportPass.enable, True )
  top2 = VerilogTranslationImportPass()( top2 )

  top2.apply( DefaultPassGroup() )
  top2.sim_reset()

  for i in range(10):
    top2.sim_tick()

  top2.finalize()
